# General
APPNAME = "SWANe"
app_acronym = "Standardized Workflow for Advanced Neuroimaging in Epilepsy"
EXECBUTTONTEXT = "Execute " + APPNAME + " Workflow"
EXECBUTTONTEXT_disabled_tooltip = "Generate Workflow first"
EXECBUTTONTEXT_STOP = "Stop " + APPNAME + " Workflow"
GENBUTTONTEXT = "Generate " + APPNAME + " Workflow"
SUBJCONFIGBUTTONTEXT = "Workflow preferences"
INFOCHAR = "\u24d8"
WF_DIR_SUFFIX = "_nipype"

# Main
main_multiple_instances_error = (
    "Another instance of " + APPNAME + " is already running!"
)

# Main Window
mainwindow_choose_working_dir = (
    "Choose the main working directory before start to use this application"
)
mainwindow_working_dir_space_error = (
    "Blank spaces are not allowed in main working dir name or in its parent folder name"
)
mainwindow_choose_working_dir_title = "Select the main working directory"
mainwindow_select_subj_folder = "Select a subject folder"
mainwindow_subj_folder_outside_workingdir_error = (
    "The selected folder is not in " + APPNAME + " main working directory!"
)
mainwindow_subj_folder_with_blank_spaces_error = (
    "The selected folder name contains blank spaces!"
)
mainwindow_subj_already_loaded_error = (
    "The selected subject was already loaded in " + APPNAME + "!"
)
mainwindow_invalid_folder_error = (
    "The selected folder does not contains valid subject data!"
)
mainwindow_force_dir_update = (
    "If you are SURE you selected a subject folder, " + APPNAME + " can try to update "
    "it.\nDo you want to update selected subject folder?"
)
mainwindow_max_subj_error = "Max subject tab limit reached!"
mainwindow_update_available = (
    "New version available (%s). We recommend you to update "
    + APPNAME
    + " with:<br><b><center>pip3 install --upgrade swane"
)
mainwindow_new_subj_name = "Write the name of the new subject:"
mainwindow_new_subj_title = "New subject"
mainwindow_new_subj_created = "New subject created: "
mainwindow_new_subj_name_error = "Invalid name: "
mainwindow_subj_exists_error = "This subject already exists: "
mainwindow_home_tab_name = "Home"
mainwindow_wf_executing_error_1 = "Cannot close a subject during workflow execution!"
mainwindow_wf_executing_error_2 = (
    "Cannot close " + APPNAME + " during workflow execution!"
)
mainwindow_home_label1 = "Welcome to " + APPNAME + "!"
mainwindow_home_label2 = (
    APPNAME
    + " ("
    + app_acronym
    + ") is a graphic tools for modular neuroimaging processing. "
    "With "
    + APPNAME
    + " you can easily import and organize DICOM files from multiple sources, "
    "generate a pipeline based on available imaging modalities and export results in a "
    "multimodal scene."
)
mainwindow_home_label3 = (
    APPNAME
    + " does NOT implement processing software but integrates in a user-friendly "
    "interface many external applications, so make sure the check the following dependencies."
)
mainwindow_home_label4 = APPNAME + " is not meant for clinical use!\n"
mainwindow_home_label5 = "\nExternal mandatory dependencies:"
mainwindow_home_label6 = "\nExternal recommended dependencies:"
mainwindow_home_label7 = "\nExternal optional dependencies:"

mainwindow_dep_slicer_src = "Searching Slicer installation..."
mainwindow_pref_disabled_error = "Preferences disabled during workflow execution!"
aboutwindow_wiki_dependencylist = "Link to <a href='https://github.com/LICE-dev/swane/wiki/04-Dependencies-Guides'>dependencies list</a>"
aboutwindow_wiki_changelog = "Link to versions <a href='https://github.com/LICE-dev/swane/wiki/11-Changelog'>changelog</a>"

mainwindow_chatgpt_title = "SWANe Assistant"
mainwindow_chatgpt_tooltip = "Open ChatGPT SWANe Assistant in your browser"

# Menu
menu_load_subj = "&Load subject..."
menu_load_subj_tip = "Load subject data from the main working directory"
menu_new_subj = "&New subject..."
menu_new_subj_tip = "Add a new subject in the main working directory"
menu_exit = "E&xit " + APPNAME
menu_pref = "Application &Settings..."
menu_pref_tip = "Edit " + APPNAME + " settings"
menu_wf_pref = "Default Workflow &Preferences..."
menu_shutdown_pref = "Shutdown at Workflow end"
menu_about = "&About " + APPNAME + "..."
menu_file_name = "File"
menu_tools_name = "Tools"
menu_help_name = "Help"
menu_auto_import = "Automatic series classification"

# Subject Tab
subj_tab_data_tab_name = "Data load"
subj_tab_wf_tab_name = "Workflow execution"
subj_tab_results_tab_name = "Results export"
subj_tab_wf_executed = APPNAME + " Workflow executed!"
subj_tab_wf_executed_with_error = APPNAME + " Workflow finished. Error occurred!"
subj_tab_import_button = "Import"
subj_tab_clear_button = "Clear"
subj_tab_scan_dicom_button = "Scan DICOM folder"
subj_tab_selected_series_error = "No series was selected"
subj_tab_wrong_type_check = "Do you want to continue importing?"
subj_tab_wrong_type_check_msg = "You selected %s images while %s images were expected."
subj_tab_wrong_max_vols_check_msg = (
    "The series you selected contains %d volumes while a maximum of %d were expected."
)
subj_tab_wrong_min_vols_check_msg = (
    "The series you selected contains %d volumes while a minimum of %d were expected."
)
subj_tab_import_copy_error_msg = "Error copying selected series files."
subj_tab_import_folder_not_empy = "Images are already loaded for this series."
subj_tab_found_series_type = "{series_description} could be associated with input {data_label}.\nDou you want to import it?"
subj_tab_dicom_copy = "Copying DICOM files in subject folder..."
subj_tab_dicom_check = "Verifying subject folder..."
subj_tab_dicom_scan = "Scanning folder for primary non derived DICOM images..."
subj_tab_subj_loading = "Checking subject DICOM folders..."
subj_tab_select_dicom_folder = "Select a folder to scan for DICOM files"
subj_tab_no_dicom_error = "No DICOM file in "
subj_tab_multi_subj_error = "Dicom file from more than one subject in "
subj_tab_multi_exam_error = "DICOM file from more than one examination in "
subj_tab_multi_series_error = "DICOM file from more than one series in "
subj_tab_missing_fsl_error = "FSL is required to generate " + APPNAME + " Workflow!"
subj_tab_wf_gen_error = "Error generating the Workflow!"
subj_tab_old_wf_found = (
    "This subject has already been analyzed by "
    + APPNAME
    + """. Do you want to resume the previous analysis? If you want to delete all
previous analyses and start over press NO, otherwise press YES"""
)
subj_tab_old_wf_resume = "Resume execution"
subj_tab_old_wf_reset = "New execution"
subj_tab_old_fs_found = "An existing FreeSurfer folder was detected. Do you want to keep or delete the existing folder?"
subj_tab_old_fs_resume = "Keep folder"
subj_tab_old_fs_reset = "Delete folder"
subj_tab_wf_stop = "Do you REALLY want to stop " + APPNAME + " Workflow execution?"
subj_tab_generate_scene_button = "Create/Update Slicer scene " + INFOCHAR
subj_tab_generate_scene_button_tooltip = (
    "In case of a new analysis remember to update the Slicer scene"
)
subj_tab_generate_scene_button_disabled_tooltip = "Slicer not detected"
subj_tab_load_scene_button = "Visualize scene into Slicer"
subj_tab_load_scene_button_tooltip = "Create a Slicer scene first"
subj_tab_open_results_directory = "Open results folder"
subj_tab_exporting_start = (
    "Exporting results into Slicer scene...\nLoading Slicer environment"
)
subj_tab_exporting_prefix = "Exporting results into Slicer scene...\n"
subj_tab_dicom_clearing = "Clearing DICOM files in: "
subj_tab_wf_insufficient_resources = (
    "Insufficient system resources (RAM or CPU) to execute workflows"
)
subj_tab_wf_invalid_signal = (
    "Signaling error, workflow will be stopped. Report this error."
)
subj_tab_wf_error_oom_gpu = "Process killed: Out of Memory (GPU). Try to reduce GPU process limit in performance preferences"
subj_tab_wf_error_oom = "Process killed (possible Out of Memory). Try to reduce CPU core limit in performance preferences"
subj_tab_wf_error_terminated = (
    "Process terminated. This can occur if user manually terminates a running process."
)
subj_tab_tabtooltip_exec_disabled_series = (
    "A 3D T1w series is required to enable workflow execution"
)
subj_tab_tabtooltip_exec_disabled_dependency = (
    "Mandatory dependencies are required to enable workflow execution"
)
subj_tab_tabtooltip_result_disabled = "Complete a workflow first"
subj_tab_tabtooltip_data_disabled = (
    "Cannot change subject data during workflow execution"
)
subj_tab_unsupported_files = (
    APPNAME
    + " works on primary non derived DICOM images.\nFolder contains ImageType combinations that are not accepted: {}"
)

# Wf Preference Window
wf_pref_window_title_user = " - Workflow preferences"

# Preference Window
pref_window_save_button = "Save preferences"
pref_window_save_restart_button = (
    "Save preferences (" + APPNAME + " will close and restart)"
)
pref_window_discard_button = "Discard changes"
pref_window_dir_error = "Directory does not exists!"
pref_window_file_error = "File does not exists!"

pref_window_reset_global_button = "Reset workflow settings to default"
pref_window_reset_global_box = (
    "Do you really want to reset global workflow settings to application default?"
)
pref_window_reset_subj_button = "Apply default workflow settings"
pref_window_reset_subj_box = (
    "Do you really want to apply default workflow settings to this subject?"
)

pref_window_mail_test_button = "Test email settings"
pref_window_mail_test_hint = "By clicking this button a test mail wil be send using the mail settings preferences"
pref_window_mail_test_fail = (
    "An error occurred, check " + APPNAME + " mail configuration"
)
pref_window_mail_test_success = "Mail sent succesfully, check in your inbox"

# Workflow
check_dep_generic_error = "Dependency check error"
check_dep_dcm2niix_error = (
    "dcm2niix not detected (<a href='https://github.com/rordenlab/dcm2niix#Install"
    "'>installation info</a>)"
)
check_dep_dcm2niix_found = "dcm2niix detected (%s)"
check_dep_fsl_error = (
    "FSL not detected (<a href='https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FslInstallation"
    "'>installation info</a>)"
)
check_dep_fsl_wrong_version = (
    "FSL version outdated (found %s, required %s). Please "
    "<a href='https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FslInstallation'>update</a>"
)
check_dep_fsl_no_locale = (
    "FSL detected, but locale en_US.utf8 not available. Use: sudo locale-gen en_US.utf8"
)
check_dep_fsl_found = "FSL detected (%s)"
check_dep_fs_found = "FreeSurfer detected (%s)"
check_dep_fs_error1 = (
    "FreeSurfer not detected (<a href='https://surfer.nmr.mgh.harvard.edu/fswiki/DownloadAndInstall"
    "'>installation info</a>)"
)
check_dep_fs_error2 = "FreeSurfer detected (%s), but without environment configuration"
check_dep_fs_error3 = (
    "FreeSurfer detected (%s). Matlab Runtime is not installed (<a "
    "href='https://surfer.nmr.mgh.harvard.edu/fswiki/MatlabRuntime'>installation instruction</a>)"
)
check_dep_fs_error4 = (
    "FreeSurfer detected (%s). License key missing (<a "
    "href='https://surfer.nmr.mgh.harvard.edu/registration.html'>registration instruction</a>)"
)
check_dep_fs_wrong_version = (
    "FreeSurfer version outdated (found %s, required %s). Please "
    "<a href='https://surfer.nmr.mgh.harvard.edu/fswiki/DownloadAndInstall'>update</a>"
)
check_dep_fs_no_tcsh = "FreeSurfer detected (%s) but tcsh is not installed."
check_dep_graph_error = "Graphviz not detected (<a href='https://graphviz.org/download/'>Installation info</a>)"
check_dep_graph_found = "Graphviz detected"
check_dep_slicer_error1 = (
    "Slicer not detected (<a href='https://slicer.readthedocs.io/en/latest/user_guide"
    "/getting_started.html#installing-3d-slicer/'>Installation info</a>)"
)
check_dep_slicer_error2 = (
    "Slicer detected but without SlicerFreeSurfer extension (<a "
    "href='https://slicer.readthedocs.io/en/latest/user_guide/extensions_manager.html?highlight"
    "=extension%20manager'>Extensions Manager info</a>)"
)
check_dep_slicer_wrong_version = (
    "Slicer version outdated (found %s, required %s). Please "
    "<a href='https://slicer.readthedocs.io/en/latest/user_guide'>update</a>"
)
check_dep_slicer_found = "Slicer detected (%s)"

fsl_python_error = (
    APPNAME
    + " has been executed using fsl Python instead of system Python.\nThis may depend "
    "on a conflict in FSL(>=6.0.6) and FreeSurfer(<=7.3.2) configurations in your %s file that "
    "impacts on correct functioning of "
    + APPNAME
    + " and maybe other applications.\n"
    + APPNAME
    + " can try to fix your configuration file or to restart with system Python interpreter. Otherwise"
    " you can exit "
    + APPNAME
    + " and fix your configuration manually adding this line to your "
    "configuration file:"
)
fsl_python_error_fix = "Fix error and Restart"
fsl_python_error_restart = "Restart with system Python"
fsl_python_error_exit = "Copy fix line and Exit"
generic_shell_file = "your shell configuration"

# Nodes
node_names = {}
node_names["CustomDcm2niix"] = "nifti conversion"
node_names["RobustFOV"] = "neck removal"
node_names["ForceOrient"] = "standard orientation"
node_names["BET"] = "scalp removal"
node_names["FLIRT"] = "linear registration"
node_names["ApplyXFM"] = "linear transformation"
node_names["FNIRT"] = "nonlinear registration"
node_names["ApplyWarp"] = "nonlinear transformation"
node_names["InvWarp"] = "inverse transformation"
node_names["DataSink"] = "saving"
node_names["ApplyMask"] = "masking"
node_names["EddyCorrect"] = "eddy current correction (old)"
node_names["CustomEddy"] = "eddy current correction"
node_names["GenEddyFiles"] = "eddy current correction preparation"
node_names["CustomBEDPOSTX5"] = "diffusion bayesian estimation"
node_names["RandomSeedGenerator"] = "random seeds generation"
node_names["CustomProbTrackX2"] = "probabilistic tractography"
node_names["SumMultiTracks"] = "Parallel tractography merging"
node_names["ReconAll"] = "Freesurfer recon-all"
node_names["CustomLabel2Vol"] = "linear transformation"
node_names["SegmentHA"] = "hippocampal segmentation"
node_names["MCFLIRT"] = "motion correction"
node_names["CustomSliceTimer"] = "slice timing correction"
node_names["SUSAN"] = "noise reduction"
node_names["FMRIGenSpec"] = "functional model generation"
node_names["ArtifactDetect"] = "outliers detection"
node_names["SpecifyModel"] = "functional model application"
node_names["Level1Design"] = "FEAT files generation"
node_names["FEATModel"] = "design file generation"
node_names["FILMGLS"] = "General-Linear-Model estimation"
node_names["SmoothEstimate"] = "smoothness estimation"
node_names["Cluster"] = "cluster extraction"
node_names["SampleToSurface"] = "surface projection"
node_names["FAST"] = "Tissue segmentation"
node_names["FLAT1OutliersMask"] = "outliers mask generation"



# slicerpython script for module checking
# Warning: slicer library is not required beacuse the script is executed in Slicer environment

import sys
import os
import subprocess


def load_anat(scene_dir: str, volume_name: str, color_node_ID: str = None):
    """
    Creates a scalar node from a NIFTI file.

    Parameters
    ----------
    scene_dir : str
        The scene directory.
    volume_name : str
        The NIFTI file name (without extension).
    color_node_ID : str, optional
        The colorNodeID to apply to Slicer node. The default is None.

    Returns
    -------
    node : MRMLCore.vtkMRMLScalarVolumeNode
        The scalar node generated from the NIFTI file.

    """

    file = os.path.join(scene_dir, volume_name + ".nii.gz")
    node = None
    if os.path.exists(file):
        try:
            print("SLICERLOADER: Loading " + volume_name)
            node = slicer.util.loadVolume(file)
        except:
            pass

        if node is not None and color_node_ID is not None:
            node.GetDisplayNode().SetAndObserveColorNodeID(color_node_ID)
            return

    return node


def lesion_segment(scene_dir: str):
    """
    Prepares an empty segment for optional manual segmentation.

    Parameters
    ----------
    scene_dir : str
        The scene directory.

    Returns
    -------
    None.

    """

    file = os.path.join(scene_dir, "seg_lesions.seg.nrrd")
    if os.path.exists(file):
        print("SLICERLOADER: Loading existing lesion segment")
        slicer.util.loadSegmentation(file)
    else:
        print("SLICERLOADER: Creating lesion segment")
        seg_node = slicer.mrmlScene.AddNewNodeByClass(
            "vtkMRMLSegmentationNode", "seg_lesions"
        )
        seg_node.CreateDefaultDisplayNodes()
        seg_node.GetSegmentation().AddEmptySegment("Lesion", "Lesion", [1, 0, 0])
        seg_node.CreateClosedSurfaceRepresentation()
        my_storage_node = seg_node.CreateDefaultStorageNode()
        my_storage_node.SetFileName(file)
        my_storage_node.WriteData(seg_node)


def load_freesurfer_surf(scene_dir: str, node_name: str, ref_node):
    """
    Loads the FreeSurfer surface.

    Parameters
    ----------
    scene_dir : str
        The scene directory.
    node_name : str
        The surface file.
    ref_node : MRMLCore.vtkMRMLScalarVolumeNode
        The reference node for surface positioning.

    Returns
    -------
    #TODO specificare tipo di surf_node
    surf_node : TYPE
        The slicer node with loaded surface.

    """

    file = os.path.join(scene_dir, node_name)
    surf_node = None

    if os.path.exists(file):
        try:
            print("SLICERLOADER: Loading surface " + node_name)
            loaded_nodes = vtk.vtkCollection()
            surf_node = slicer.app.ioManager().loadNodes(
                "FreeSurfer model",
                {"fileName": file, "referenceVolumeID": ref_node.GetID()},
                loaded_nodes,
            )
            if surf_node:
                surf_node = list(loaded_nodes)[0]
                surf_node.GetDisplayNode().SetColor(0.82, 0.82, 0.82)
        except:
            pass

    return surf_node


def load_freesurfer_overlay(scene_dir: str, node_name: str, surf_node):
    """
    Applies an overlay on the FreeSurfer surface.

    Parameters
    ----------
    scene_dir : str
        The scene directory.
    node_name : str
        The surface file.
    #TODO specificare tipo di surf_node
    surf_node : TYPE
        The slicer node with loaded surface.

    Returns
    -------
    None.

    """

    file = os.path.join(scene_dir, node_name)
    if os.path.exists(file):
        try:
            print("SLICERLOADER: Loading surface overlay " + node_name)
            overlay_node = slicer.util.loadNodeFromFile(
                file, "FreeSurferScalarOverlayFile", {"modelNodeId": surf_node.GetID()}
            )
            overlay_node.GetDisplayNode().SetAndObserveColorNodeID(
                "vtkMRMLColorTableNodeFileColdToHotRainbow.txt"
            )
            overlay_node.GetDisplayNode().SetScalarVisibility(False)
        except:
            pass


def load_freesurfer_segmentation_file(seg_file: str):
    """
    Loads the FreeSurfer segmentation file.

    Parameters
    ----------
    seg_file : str
        The segmentation file name.

    Returns
    -------
    None.

    """

    if os.path.exists(seg_file):
        try:
            # OLD COMMON METHOD
            # slicer.util.loadNodeFromFile(seg_file, 'FreeSurferSegmentationFile')

            # WARNING: SlicerFreesurfer plugin bug causing Crash on ubuntu with the new method.
            # Preserving old method to ensure functionality with older version of the plugin.
            if "linux" in sys.platform:
                slicer.util.loadVolume(
                    seg_file,
                    {"labelmap": True, "colorNodeID": "vtkMRMLColorTableNodeFile"},
                )
            else:
                slicer.util.getModuleLogic(
                    "FreeSurferImporter"
                ).LoadFreeSurferSegmentation(seg_file)

        except:
            pass


def load_freesurfer(scene_dir: str, ref_node):
    """
    Loads the FreeSurfer output.

    Parameters
    ----------
    scene_dir : str
        The scene directory.
    ref_node : MRMLCore.vtkMRMLScalarVolumeNode
        The reference node for surface positioning.

    Returns
    -------
    None.

    """

    seg_list = [
        "r-aparc_aseg.mgz",
        "segmentHA/lh.hippoAmygLabels.mgz",
        "segmentHA/rh.hippoAmygLabels.mgz",
    ]
    for file in seg_list:
        seg_file = os.path.join(scene_dir, file)
        load_freesurfer_segmentation_file(seg_file)

    lh_pial = load_freesurfer_surf(scene_dir, "lh.pial", ref_node)
    rh_pial = load_freesurfer_surf(scene_dir, "rh.pial", ref_node)

    surfs = ["lh.white", "rh.white"]
    for surf in surfs:
        load_freesurfer_surf(scene_dir, surf, ref_node)

    overlays = [
        "pet_surf",
        "pet_ai_surf",
        "pet_zscore_surf",
        "asl_surf",
        "asl_ai_surf",
        "asl_zscore_surf",
    ]
    for overlay in overlays:
        load_freesurfer_overlay(scene_dir, overlay + "_lh.mgz", lh_pial)
        load_freesurfer_overlay(scene_dir, overlay + "_rh.mgz", rh_pial)


def load_vein(scene_dir: str, remove_vein: bool = False, vein_thresold: float = 97.5):
    """
    Loads the veins files and creates their 3d model.

    Parameters
    ----------
    scene_dir : str
        The scene directory.
    remove_vein : bool, optional
        True if the model must remove the original veins images. The default is False.
    vein_thresold : float, optional
        The threshold for the vein detection in 3D Slicer. The default is 97.5.

    Returns
    -------
    None.

    """

    vein_volume_name = "r-veins_inskull"
    vein_node = load_anat(scene_dir, vein_volume_name)
    if vein_node is None:
        return
    print("SLICERLOADER: Creating 3D model: Veins")

    try:
        command = (
            "fslstats "
            + os.path.join(scene_dir, vein_volume_name + ".nii.gz")
            + " -P "
            + str(vein_threshold)
        )
        output = subprocess.run(
            command, shell=True, stdout=subprocess.PIPE
        ).stdout.decode("utf-8")
        thr = float(output)
    except:
        thr = 6

    vein_model = slicer.vtkMRMLModelNode()
    slicer.mrmlScene.AddNode(vein_model)

    parameters = {
        "InputVolume": vein_node.GetID(),
        "Threshold": thr,
        "OutputGeometry": vein_model.GetID(),
    }

    gray_maker = slicer.modules.grayscalemodelmaker
    slicer.cli.runSync(gray_maker, None, parameters)
    vein_model.GetDisplayNode().SetColor(0, 0, 1)
    vein_model.SetName("Veins")
    my_storage_node = vein_model.CreateDefaultStorageNode()
    my_storage_node.SetFileName(os.path.join(scene_dir, "veins.vtk"))
    my_storage_node.WriteData(vein_model)

    if remove_vein:
        slicer.mrmlScene.RemoveNode(vein_node)


def tract_model(
    segmentation_node, dti_dir: str, tract: [], side: str, dti_threshold: float = 0.0035
):
    """
    Creates the 3d model of a tract.

    Parameters
    ----------
    #TODO verificare tipo
    segmentation_node : TYPE
        DESCRIPTION.
    dti_dir : str
        The DTI directory.
    tract : []
        The array with the tract information.
    side : str
        The side of the tract:
            - LH for left side
            - RH for right side
    dti_threshold : float
        The threshold for the dti tractography

    Returns
    -------
    None.

    """

    tract_file = os.path.join(dti_dir, "r-" + tract["name"] + "_" + side + ".nii.gz")
    if not os.path.exists(tract_file):
        return

    waytotal_file = os.path.join(
        dti_dir, "r-" + tract["name"] + "_" + side + "_waytotal"
    )
    waytotal = 0
    if os.path.exists(waytotal_file):
        try:
            with open(waytotal_file, "r") as file:
                for line in file.readlines():
                    waytotal = int(line)
        except:
            pass

    if waytotal > 0:
        thr = waytotal * dti_threshold
    else:
        thr = tract["thr"]

    tract_node = slicer.util.loadVolume(tract_file)
    segmentation_node.SetReferenceImageGeometryParameterFromVolumeNode(tract_node)

    # Create temporary segment editor to get access to effects
    segment_editor_widget = slicer.qMRMLSegmentEditorWidget()
    segment_editor_widget.setMRMLScene(slicer.mrmlScene)
    segment_editor_node = slicer.mrmlScene.AddNewNodeByClass("vtkMRMLSegmentEditorNode")
    segment_editor_widget.setMRMLSegmentEditorNode(segment_editor_node)
    segment_editor_widget.setSegmentationNode(segmentation_node)
    segment_editor_widget.setSourceVolumeNode(tract_node)

    # Create segment
    tract_segment_id = segmentation_node.GetSegmentation().AddEmptySegment(
        tract["name"], tract["name"], tract["color"]
    )
    segment_editor_node.SetSelectedSegmentID(tract_segment_id)

    # Fill by thresholding
    segment_editor_widget.setActiveEffectByName("Threshold")
    effect = segment_editor_widget.activeEffect()
    effect.setParameter("MinimumThreshold", thr)
    effect.self().onApply()

    # Delete temporary segment editor
    segment_editor_widget = None
    slicer.mrmlScene.RemoveNode(segment_editor_node)
    slicer.mrmlScene.RemoveNode(tract_node)


def load_fmri(scene_dir: str):
    """
    Loads the fMRI results.

    Parameters
    ----------
    scene_dir : str
        The scene directory.

    Returns
    -------
    None.

    """

    fmri_path = os.path.join(scene_dir, "fMRI")
    if not os.path.exists(fmri_path):
        return
    print("SLICERLOADER: Loading fMRI")
    for file in os.listdir(fmri_path):
        if file.endswith(".nii.gz"):
            func_node = load_anat(
                fmri_path,
                file.replace(".nii.gz", ""),
                "vtkMRMLPETProceduralColorNodePET-Rainbow2",
            )


def main_tract(dti_dir: str, scene_dir: str):
    sides = ["rh", "lh"]
    tracts = [
        {"name": "cst", "thr": "500", "color": [0, 1, 0]},
        {"name": "af", "thr": "1500", "color": [1, 0, 1]},
        {"name": "or", "thr": "500", "color": [1, 1, 0]},
    ]

    print("SLICERLOADER: Creating DTI tracts 3D models (some minutes!)")

    for side in sides:
        # Create segmentation
        segmentation_node = slicer.mrmlScene.AddNewNodeByClass(
            "vtkMRMLSegmentationNode", "tracts_" + side
        )
        segmentation_node.CreateDefaultDisplayNodes()  # only needed for display

        # DTI Threshold preferences parsing
        if sys.argv[2] is not None:
            try:
                dti_threshold = float(sys.argv[2])
            except:
                dti_threshold = 0.0035
        for tract in tracts:
            tract_model(
                segmentation_node, dti_dir, tract, side, dti_threshold=dti_threshold
            )

        segmentation_node.CreateClosedSurfaceRepresentation()
        my_storage_node = segmentation_node.CreateDefaultStorageNode()
        my_storage_node.SetFileName(
            os.path.join(scene_dir, "tracts_" + side + ".seg.nrrd")
        )
        my_storage_node.WriteData(segmentation_node)


###############################################################################

# slicerpython execution code
slicer.mrmlScene.Clear(0)
results_folder = os.getcwd()

if not os.path.isdir(results_folder):
    print("SLICERLOADER: Results folder not found")
else:
    refNode = load_anat(results_folder, "ref")
    if refNode is not None:

        dtiDir = os.path.join(results_folder, "dti")
        if os.path.isdir(dtiDir):
            main_tract(dtiDir, results_folder)

        # lesion_segment(sceneDir)

        baseList = [
            "ref_brain",
            "r-flair_brain",
            "r-flair",
            "r-mdc_brain",
            "r-mdc",
            "r-FA",
            "r-flair2d_tra_brain",
            "r-flair2d_cor_brain",
            "r-flair2d_sag_brain",
            "r-t2_cor_brain",
            "r-t2_cor",
            "r-binary_flair",
            "r-junction_z",
            "r-extension_z",
        ]

        for volume in baseList:
            load_anat(results_folder, volume)

        colorList = [
            ("r-asl_ai", "vtkMRMLColorTableNodeFileDivergingBlueRed.txt"),
            ("r-pet_ai", "vtkMRMLColorTableNodeFileDivergingBlueRed.txt"),
            ("r-pet", "vtkMRMLColorTableNodeFileColdToHotRainbow.txt"),
            ("r-pet_zscore", "vtkMRMLColorTableNodeFileColdToHotRainbow.txt"),
            ("r-asl", "vtkMRMLColorTableNodeFileColdToHotRainbow.txt"),
            ("r-asl_zscore", "vtkMRMLColorTableNodeFileColdToHotRainbow.txt"),
        ]

        for volume in colorList:
            load_anat(results_folder, volume[0], volume[1])

        load_fmri(results_folder)

        # Vein Threshold preferences parsing
        if sys.argv[3] is not None:
            try:
                vein_threshold = float(sys.argv[3])
            except:
                vein_threshold = 97.5

        load_vein(results_folder, vein_thresold=vein_threshold)

        load_freesurfer(results_folder, refNode)

        ext = "mrb"

        # TODO valutare
        # Saving in MRML doesn't work well, disable extension choice for now
        # if sys.argv[4] is not None and sys.argv[1] == "mrml":
        #     ext = "mrml"

        print("SLICERLOADER: Saving multimodale scene (some minutes)")
        slicer.util.saveScene(os.path.join(results_folder, "scene." + ext))

sys.exit(0)



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nibabel import load, save
from nibabel.processing import conform

from os.path import abspath
import os
import shutil

from nipype.interfaces.base import (
    traits,
    BaseInterface,
    BaseInterfaceInputSpec,
    TraitedSpec,
    File,
    isdefined,
)


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterfaceInputSpec)  -*-
class CropFovInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="the input image")
    max_dim = traits.Int(mandatory=True)
    out_file = File(desc="the output image")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class CropFovOutputSpec(TraitedSpec):
    out_file = File(desc="the output image")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterface)  -*-
class CropFov(BaseInterface):
    """
    If FOV exceeds 250mm, crop the borders.

    """

    input_spec = CropFovInputSpec
    output_spec = CropFovOutputSpec

    def _run_interface(self, runtime):
        self.inputs.out_file = self._gen_outfilename()

        # Calculate current fov in mm
        img = load(self.inputs.in_file)
        dim1, dim2, dim3 = img.header.get_data_shape()
        vox1, vox2, vox3 = img.header.get_zooms()
        fov = [dim1 * vox1, dim2 * vox2, dim3 * vox3]
        current_max_dim = max(fov)

        if current_max_dim <= self.inputs.max_dim:
            shutil.copyfile(self.inputs.in_file, self.inputs.out_file)
        else:

            rescale_factor = (self.inputs.max_dim - 1) / current_max_dim

            t_dim1 = (
                dim1 if fov[0] <= self.inputs.max_dim else int(dim1 * rescale_factor)
            )
            t_dim2 = (
                dim2 if fov[1] <= self.inputs.max_dim else int(dim2 * rescale_factor)
            )
            t_dim3 = (
                dim3 if fov[2] <= self.inputs.max_dim else int(dim3 * rescale_factor)
            )

            scaled_img = conform(img, (t_dim1, t_dim2, t_dim3), img.header.get_zooms())

            save(scaled_img, self.inputs.out_file)

        return runtime

    def _gen_outfilename(self):
        out_file = self.inputs.out_file
        if not isdefined(out_file) and isdefined(self.inputs.in_file):
            out_file = "cropped_" + os.path.basename(self.inputs.in_file)
        return abspath(out_file)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self._gen_outfilename()
        return outputs



from PySide6.QtWidgets import QProgressDialog
from PySide6.QtCore import Qt


class PersistentProgressDialog(QProgressDialog):
    """
    Custom implementation of PySide QProgressDialog to define a non-closable window.

    """

    def __init__(self, label_text, minimum, maximum, parent=None):
        super(PersistentProgressDialog, self).__init__(
            label_text, None, minimum, maximum, parent, Qt.WindowFlags()
        )
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowCloseButtonHint)
        self.setWindowModality(Qt.WindowModal)
        self.setMinimumDuration(0)

    def keyPressEvent(self, event):
        if event.key() != Qt.Key_Escape:
            super(PersistentProgressDialog, self).keyPressEvent(event)

    def increase_value(self, x, maximum=0):
        if self.maximum() == 0 and maximum > 0:
            self.setMaximum(maximum)
        if not self.isVisible():
            self.show()
        if self.value() == 0 and self.maximum() > 1:
            x = x + 1
        self.setValue(self.value() + x)

    def closeEvent(self, event):
        event.ignore()



from swane.config.config_enums import InputTypes
from enum import Enum


class PreferenceEntry:
    """
    Contains information to manage loading, saving and visualization of a preference
    """

    input_type = InputTypes.TEXT
    label = ""
    default = None
    tooltip = ""
    range = None
    decimals = None
    dependency = None
    dependency_fail_tooltip = None
    pref_requirement = None
    pref_requirement_fail_tooltip = None
    input_requirement = None
    input_requirement_fail_tooltip = None
    restart = False
    validate_on_change = False
    informative_text: dict = None
    box_text = None
    hidden = False
    value_enum = None
    default_at_startup = False

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            if hasattr(self, key) and PreferenceEntry.check_type(key, value):
                setattr(self, key, value)

    @staticmethod
    def check_type(key, value):
        types = {
            "input_type": InputTypes,
            "label": str,
            "tooltip": str,
            "range": list,
            "decimals": int,
            "dependency": str,
            "dependency_fail_tooltip": str,
            "pref_requirement": dict,
            "pref_requirement_fail_tooltip": str,
            "input_requirement": list,
            "input_requirement_fail_tooltip": str,
            "restart": bool,
            "validate_on_change": bool,
            "informative_text": dict,
            "box_text": str,
            "hidden": bool,
            # "value_enum": Enum
        }
        if key in types:
            if types[key] == Enum:
                return isinstance(value, Enum)
            return type(value) is types[key]
        return True



from nipype.interfaces.freesurfer import ReconAll
from nipype.interfaces.fsl import BinaryMaths
from multiprocessing import cpu_count
from nipype.pipeline.engine import Node
from math import trunc
from swane.nipype_pipeline.nodes.utils import getn
from swane.nipype_pipeline.engine.CustomWorkflow import CustomWorkflow
from swane.nipype_pipeline.nodes.SegmentHA import SegmentHA
from swane.nipype_pipeline.nodes.CustomLabel2Vol import CustomLabel2Vol
from swane.nipype_pipeline.nodes.ThrROI import ThrROI
from swane.config.config_enums import CORE_LIMIT
from nipype.interfaces.utility import IdentityInterface

FS_DIR = "FS"


def freesurfer_workflow(
    name: str,
    is_hippo_amyg_labels: bool,
    base_dir: str = "/",
    max_cpu: int = 0,
    multicore_node_limit: CORE_LIMIT = CORE_LIMIT.SOFT_CAP,
) -> CustomWorkflow:
    """
    Freesurfer cortical reconstruction, white matter ROI, basal ganglia and thalami ROI.
    If needed, segmentation of the hippocampal substructures and the nuclei of the amygdala.

    Parameters
    ----------
    name : str
        The workflow name.
    is_hippo_amyg_labels : bool
        Enable segmentation of the hippocampal substructures and the nuclei of the amygdala.
    base_dir : path, optional
        The base directory path relative to parent workflow. The default is "/".
    max_cpu : int, optional
        If greater than 0, limit the core usage of bedpostx. The default is 0.
    multicore_node_limit: CORE_LIMIT, optional
        Preference for bedpostX core usage. The default il CORE_LIMIT.SOFT_CAP

    Input Node Fields
    ----------
    ref : path
        T13D reference file.
    subjects_dir : path
        Directory for Freesurfer analysis.

    Returns
    -------
    workflow : CustomWorkflow
        The Freesurfer workflow.

    Output Node Fields
    ----------
    subject_id : string
        Subject name for Freesurfer (defined as FS_DIR="FS").
    subjects_dir : path
        Directory for Freesurfer analysis.
    bgROI : path
        Binary ROI for basal ganglia and thalamus.
    wmROI : path
        Binary ROI for cerebral white matter.
    pial : list of strings
        Gray matter/pia mater rh and lh surfaces.
    white : list of strings
        White/gray matter rh and lh surfaces.
    vol_label_file : path
        Aparc parcellation projected into aseg volume in reference space.
    vol_label_file_nii : path
        Aparc parcellation projected into aseg volume in reference space and nifti format.
    lh_hippoAmygLabels : path
        Left side labels from segmentation of the hippocampal substructures and the nuclei of the amygdala.
    rh_hippoAmygLabels : path
        Right side labels from segmentation of the hippocampal substructures and the nuclei of the amygdala.

    """

    workflow = CustomWorkflow(name=name, base_dir=base_dir)

    # Input Node
    inputnode = Node(
        IdentityInterface(fields=["ref", "subjects_dir"]), name="inputnode"
    )

    # Output Node
    outputnode = Node(
        IdentityInterface(
            fields=[
                "subject_id",
                "subjects_dir",
                "bgROI",
                "wmROI",
                "pial",
                "white",
                "vol_label_file",
                "vol_label_file_nii",
                "lh_hippoAmygLabels",
                "rh_hippoAmygLabels",
            ]
        ),
        name="outputnode",
    )

    # NODE 1: Freesurfer cortical reconstruction process
    recon_all = Node(ReconAll(), name="reconAll")
    recon_all.inputs.subject_id = FS_DIR

    # parallel option split some steps in right and left
    if max_cpu > 1:
        recon_all.inputs.parallel = True

    # openmp option apply max cpu tu some steps, resulting in twice cpu usage for rogh/left steps
    if multicore_node_limit == CORE_LIMIT.NO_LIMIT:
        # no limit
        recon_all.inputs.openmp = cpu_count()
    elif multicore_node_limit == CORE_LIMIT.SOFT_CAP:
        # for soft cap we accept that parallelized steps use each max_cpu cores, resulting in twice the setting
        recon_all.inputs.openmp = max_cpu
        recon_all.n_procs = recon_all.inputs.openmp
    else:
        # for hard cap we use half of max_cpu setting, but at least 1
        recon_all.inputs.openmp = max(trunc(max_cpu / 2), 1)
        recon_all.n_procs = recon_all.inputs.openmp * 2

    recon_all.inputs.directive = "all"
    recon_all.inputs.args = "-no-isrunning"
    workflow.add_nodes([recon_all])
    workflow.connect(inputnode, "ref", recon_all, "T1_files")
    workflow.connect(inputnode, "subjects_dir", recon_all, "subjects_dir")

    # NODE 2: Aparcaseg linear transformation in reference space
    aparaseg2Volmgz = Node(CustomLabel2Vol(), name="aparaseg2Volmgz")
    aparaseg2Volmgz.long_name = "label %s to reference space"
    aparaseg2Volmgz.inputs.vol_label_file = "./r-aparc_aseg.mgz"
    workflow.connect(recon_all, "rawavg", aparaseg2Volmgz, "template_file")
    workflow.connect(
        [(recon_all, aparaseg2Volmgz, [(("aparc_aseg", getn, 0), "reg_header")])]
    )
    workflow.connect(
        [(recon_all, aparaseg2Volmgz, [(("aparc_aseg", getn, 0), "seg_file")])]
    )
    workflow.connect(recon_all, "subjects_dir", aparaseg2Volmgz, "subjects_dir")
    workflow.connect(recon_all, "subject_id", aparaseg2Volmgz, "subject_id")

    # NODE 3: Aparcaseg conversion mgz -> nifti
    aparaseg2Volnii = Node(CustomLabel2Vol(), name="aparaseg2Volnii")
    aparaseg2Volnii.long_name = "label Nifti conversion"
    aparaseg2Volnii.inputs.vol_label_file = "r-aparc_aseg.nii.gz"
    workflow.connect(recon_all, "rawavg", aparaseg2Volnii, "template_file")
    workflow.connect(
        [(recon_all, aparaseg2Volnii, [(("aparc_aseg", getn, 0), "reg_header")])]
    )
    workflow.connect(
        [(recon_all, aparaseg2Volnii, [(("aparc_aseg", getn, 0), "seg_file")])]
    )
    workflow.connect(
        aparaseg2Volnii, "vol_label_file", outputnode, "vol_label_file_nii"
    )

    # NODE 4: Left cerebral white matter binary ROI
    lhwmROI = Node(ThrROI(), name="lhwmROI")
    lhwmROI.long_name = "Lh white matter ROI"
    lhwmROI.inputs.seg_val_min = 2
    lhwmROI.inputs.seg_val_max = 2
    lhwmROI.inputs.out_file = "lhwmROI.nii.gz"
    workflow.connect(aparaseg2Volnii, "vol_label_file", lhwmROI, "in_file")

    # NODE 5: Right cerebral white matter binary ROI
    rhwmROI = Node(ThrROI(), name="rhwmROI")
    rhwmROI.long_name = "Rh white matter ROI"
    rhwmROI.inputs.seg_val_min = 41
    rhwmROI.inputs.seg_val_max = 41
    rhwmROI.inputs.out_file = "rhwmROI.nii.gz"
    workflow.connect(aparaseg2Volnii, "vol_label_file", rhwmROI, "in_file")

    # NODE 4: Cerebral white matter binary ROI
    wmROI = Node(BinaryMaths(), name="wmROI")
    wmROI.long_name = "white matter ROI"
    wmROI.inputs.operation = "add"
    wmROI.inputs.out_file = "wmROI.nii.gz"
    workflow.connect(lhwmROI, "out_file", wmROI, "in_file")
    workflow.connect(rhwmROI, "out_file", wmROI, "operand_file")

    # NODE 7: Left basal ganglia and thalamus binary ROI
    lhbgROI = Node(ThrROI(), name="lhbgROI")
    lhbgROI.long_name = "Lh Basal ganglia ROI"
    lhbgROI.inputs.seg_val_min = 11
    lhbgROI.inputs.seg_val_max = 13
    lhbgROI.inputs.out_file = "lhbgROI.nii.gz"
    workflow.connect(aparaseg2Volnii, "vol_label_file", lhbgROI, "in_file")

    # NODE 8: Right basal ganglia and thalamus binary ROI
    rhbgROI = Node(ThrROI(), name="rhbgROI")
    rhbgROI.long_name = "Rh Basal ganglia ROI"
    rhbgROI.inputs.seg_val_min = 50
    rhbgROI.inputs.seg_val_max = 52
    rhbgROI.inputs.out_file = "rhbgROI.nii.gz"
    workflow.connect(aparaseg2Volnii, "vol_label_file", rhbgROI, "in_file")

    # NODE 9: Basal ganglia and thalami binary ROI
    bgROI = Node(BinaryMaths(), name="bgROI")
    bgROI.long_name = "Basal ganglia ROI"
    bgROI.inputs.operation = "add"
    bgROI.inputs.out_file = "bgROI.nii.gz"
    workflow.connect(lhbgROI, "out_file", bgROI, "in_file")
    workflow.connect(rhbgROI, "out_file", bgROI, "operand_file")

    workflow.connect(bgROI, "out_file", outputnode, "bgROI")
    # TODO wmROI work in progress - Not used for now. Maybe useful for SUPERFLAIR
    workflow.connect(wmROI, "out_file", outputnode, "wmROI")
    workflow.connect(recon_all, "pial", outputnode, "pial")
    workflow.connect(recon_all, "white", outputnode, "white")
    workflow.connect(recon_all, "subject_id", outputnode, "subject_id")
    workflow.connect(recon_all, "subjects_dir", outputnode, "subjects_dir")
    workflow.connect(aparaseg2Volmgz, "vol_label_file", outputnode, "vol_label_file")

    if is_hippo_amyg_labels:
        # NODE 10: Segmentation of the hippocampal substructures and the nuclei of the amygdala
        segmentHA = Node(SegmentHA(), name="segmentHA")
        if multicore_node_limit == CORE_LIMIT.NO_LIMIT:
            segmentHA.inputs.num_cpu = cpu_count()
        elif multicore_node_limit == CORE_LIMIT.SOFT_CAP:
            segmentHA.inputs.num_cpu = max_cpu
        else:
            segmentHA.inputs.num_cpu = max_cpu
            segmentHA.n_procs = segmentHA.inputs.num_cpu
        workflow.connect(recon_all, "subjects_dir", segmentHA, "subjects_dir")
        workflow.connect(recon_all, "subject_id", segmentHA, "subject_id")
        workflow.connect(
            segmentHA, "lh_hippoAmygLabels", outputnode, "lh_hippoAmygLabels"
        )
        workflow.connect(
            segmentHA, "rh_hippoAmygLabels", outputnode, "rh_hippoAmygLabels"
        )

    return workflow



import os
from swane.config.ConfigManager import ConfigManager
from swane.utils.DataInputList import DataInputList


class SubjectInputState:
    def __init__(self):
        self.loaded = False
        self.volumes = 0


class SubjectInputStateList(dict[DataInputList, SubjectInputState]):
    def __init__(self, dicom_dir: str, global_config: ConfigManager):
        super().__init__()
        self.dicom_dir = dicom_dir
        for data_input in DataInputList:
            if (
                data_input.value.optional
                and not global_config.is_optional_series_enabled(data_input)
            ):
                continue
            self[data_input] = SubjectInputState()

    def is_ref_loaded(self):
        return self[DataInputList.T13D].loaded

    def get_dicom_dir(self, data_input: DataInputList):
        return os.path.join(self.dicom_dir, str(data_input))



from enum import Enum
from swane.config.PrefCategory import PrefCategory
from swane.config.config_enums import ImageModality, PLANES

FMRI_NUM = 3


class DataInput(PrefCategory):
    def __init__(
        self,
        name: str,
        label: str = "",
        tooltip: str = "",
        image_modality: ImageModality = ImageModality.RM,
        optional: bool = False,
        workflow_name: str = None,
        max_volumes: int = 1,
        min_volumes: int = 1,
    ):
        super().__init__(name, label)
        self.tooltip = tooltip
        self.image_modality = image_modality
        self.optional = optional
        self.loaded = False
        self.volumes = 0
        self.max_volumes = max_volumes
        self.min_volumes = min_volumes
        if workflow_name is None:
            self.workflow_name = self.name
        else:
            self.workflow_name = workflow_name

    def is_image_modality(self, image_modality_found: str) -> bool:
        try:
            return (
                self.image_modality == image_modality_found
                or self.image_modality.value.lower()
                == str(image_modality_found).lower()
            )
        except:
            return False


class DataInputList(Enum):
    T13D = DataInput(name="t13d", label="3D T1w")
    FLAIR3D = DataInput(name="flair3d", label="3D Flair")
    MDC = DataInput(name="mdc", label="Post-contrast 3D T1w")
    VENOUS = DataInput(
        name="venous",
        label="Venous MRA - Phase contrast",
        tooltip="If you have anatomic and venous volumes in a single sequence, load it here. Otherwise, load one of the two volume (which one is not important)",
        max_volumes=2,
    )
    VENOUS2 = DataInput(
        name=VENOUS.name + "2",
        label="Venous MRA - Second volume (optional)",
        tooltip="If you have anatomic and venous volumes in two different sequences, load the remaining volume here. Otherwise, leave this slot empty",
        workflow_name=VENOUS.name,
    )
    DTI = DataInput(
        name="dti", label="Diffusion Tensor Imaging", max_volumes=-1, min_volumes=4
    )
    ASL = DataInput(
        name="asl",
        label="Arterial Spin Labeling",
        tooltip="CBF images from an ASL sequence",
    )
    PET = DataInput(name="pet", label="PET", image_modality=ImageModality.PET)

    # An Enum usually contains EVERY variable defined in its __init__, we want to ignore some used for loop only
    _ignore_ = "DataInputList i"
    DataInputList = vars()

    for i in PLANES:
        DataInputList["FLAIR2D_%s" % i.name] = DataInput(
            name="flair2d_%s" % i.name.lower(),
            label="2D Flair %s" % i.value,
            optional=True,
        )
    T2_COR = DataInput(name="t2_cor", label="2D T2 coronal", optional=True)

    for i in range(FMRI_NUM):
        DataInputList["FMRI_%d" % i] = DataInput(
            name="fmri_%d" % i,
            label="Task fMRI - Sequence %d" % (i + 1),
            max_volumes=-1,
            min_volumes=4,
        )

    def __str__(self):
        return self.value.name



import os
import shutil

from swane.utils.DataInputList import DataInputList, ImageModality
from enum import Enum, auto
from swane.config.ConfigManager import ConfigManager
from swane.utils.SubjectInputStateList import SubjectInputStateList
from swane.utils.DependencyManager import DependencyManager
from swane.workers.DicomSearchWorker import DicomSearchWorker
from PySide6.QtCore import QThreadPool
from swane import strings
from swane.nipype_pipeline.MainWorkflow import MainWorkflow
import traceback
from threading import Thread
from swane.nipype_pipeline.workflows.freesurfer_workflow import FS_DIR
from multiprocessing import Queue
from swane.workers.WorkflowMonitorWorker import WorkflowMonitorWorker
from swane.workers.WorkflowProcess import WorkflowProcess
from swane.workers.SlicerExportWorker import SlicerExportWorker


class SubjectRet(Enum):
    FolderNotFound = auto()
    PathBlankSpaces = auto()
    FolderOutsideMain = auto()
    FolderAlreadyExists = auto()
    InvalidFolderTree = auto()
    ValidFolder = auto()
    DataInputNonEmpty = auto()
    DataInputLoading = auto()
    DataInputWarningNoDicom = auto()
    DataInputWarningMultiSubj = auto()
    DataInputWarningMultiStudy = auto()
    DataInputWarningMultiSeries = auto()
    DataInputValid = auto()
    DataImportErrorVolumesMax = auto()
    DataImportErrorVolumesMin = auto()
    DataImportErrorModality = auto()
    DataImportErrorCopy = auto()
    DataImportCompleted = auto()
    GenWfMissingRequisites = auto()
    GenWfError = auto()
    GenWfCompleted = auto()
    ExecWfResume = auto()
    ExecWfResumeFreesurfer = auto()
    ExecWfStarted = auto()
    ExecWfStopped = auto()
    ExecWfStatusError = auto()


class Subject:
    GRAPH_DIR_NAME = "graph"
    GRAPH_FILE_PREFIX = "graph_"
    GRAPH_FILE_EXT = "svg"
    GRAPH_TYPE = "colored"

    def __init__(
        self, global_config: ConfigManager, dependency_manager: DependencyManager
    ):
        self.global_config: ConfigManager = global_config
        self.folder: str = None
        self.name: str = None
        self.input_state_list: SubjectInputStateList = None
        self.config: ConfigManager = None
        self.dependency_manager: DependencyManager = dependency_manager
        self.workflow: MainWorkflow = None
        self.workflow_process: WorkflowProcess = None
        self.workflow_monitor_work: WorkflowMonitorWorker = None

    def load(self, subject_folder: str) -> SubjectRet:
        """
        Load subject information from a folder, generate subject configuration and input_state_list

        Parameters
        ----------
        subject_folder: str
            The folder to scan

        Returns
        -------
        A return code from SubjectRet

        """

        check = self.check_subject_folder(subject_folder)
        if check != SubjectRet.ValidFolder:
            return check

        self.folder = subject_folder
        self.name = os.path.basename(subject_folder)
        self.input_state_list = SubjectInputStateList(
            self.dicom_folder(), self.global_config
        )
        self.create_config(self.dependency_manager)
        return SubjectRet.ValidFolder

    def prepare_scan_dicom_folders(
        self,
    ) -> tuple[dict[DataInputList, DicomSearchWorker], int]:
        """
        Generates a DicomSearchWorker for every data input

        Parameters
        ----------

        Returns
        -------
        A tuple formed by:
            A dict wich keys are data inputs and values are the relative DicomSearchWorkers
            An int with the number of files to be scanned by that worker

        """
        dicom_scanners = {}
        total_files = 0
        for data_input in self.input_state_list:
            dicom_scanners[data_input] = self.gen_dicom_search_worker(data_input)
            total_files = total_files + dicom_scanners[data_input].get_files_len() + 1
        return dicom_scanners, total_files

    def gen_dicom_search_worker(self, data_input: DataInputList) -> DicomSearchWorker:
        """
        Generates a Worker that scan the series folder in search for DICOM files.

        Parameters
        ----------
        data_input : DataInputList
            The series folder name to check.

        Returns
        -------
        dicom_src_work : DicomSearchWorker
            The DICOM Search Worker.

        """

        src_path = self.dicom_folder(data_input)
        dicom_src_work = DicomSearchWorker(src_path)
        dicom_src_work.load_dir()

        return dicom_src_work

    def execute_scan_dicom_folders(
        self,
        dicom_scanners: dict[DataInputList, DicomSearchWorker],
        status_callback: callable = None,
        progress_callback: callable = None,
    ):
        """
        Load subject information from a folder, generate subject configuration and input_state_list

        Parameters
        ----------
        dicom_scanners: dict[DataInputList, DicomSearchWorker]
            The list of DicomSearchWorker for every data input
        status_callback: callable, optional
            The function to notify return code. Default is None
        progress_callback: callable, optional
            The function to notify scan progress. Default is None

        """
        for data_input in self.input_state_list:
            if data_input in dicom_scanners:
                self.check_input_folder_step2(
                    data_input,
                    dicom_scanners[data_input],
                    progress_callback=progress_callback,
                    status_callback=status_callback,
                )
                if status_callback is not None:
                    status_callback(data_input, SubjectRet.DataInputLoading)

    def check_input_folder(
        self,
        data_input: DataInputList,
        status_callback: callable = None,
        progress_callback: callable = None,
    ):
        """
        Checks if the series folder labelled data_input contains DICOM files.
        If PersistentProgressDialog is not None, it will be used to show the scan progress.

        Parameters
        ----------
        data_input : DataInputList
            The series folder name to check.
        status_callback: callable, optional
            The function to notify update to UI. Default is None
        progress_callback : callable, optional
            A callback function to notify progress. The default is None.

        Returns
        -------
        None.

        """

        dicom_src_work = self.gen_dicom_search_worker(data_input)
        self.check_input_folder_step2(
            data_input,
            dicom_src_work,
            status_callback=status_callback,
            progress_callback=progress_callback,
        )

    def check_input_folder_step2(
        self,
        data_input: DataInputList,
        dicom_src_work: DicomSearchWorker,
        status_callback: callable = None,
        progress_callback: callable = None,
    ):
        """
        Starts the DICOM files scan Worker into the series folder on a new thread.

        Parameters
        ----------
        data_input : DataInputList
            The series folder name to check.
        dicom_src_work : DicomSearchWorker
            The DICOM Search Worker.
        status_callback: callable
            The function to notify update to UI. Default is None
        progress_callback : callable, optional
            A callback function to notify progress. The default is None.

        Returns
        -------
        None.

        """

        dicom_src_work.signal.sig_finish.connect(
            lambda src, name=data_input, callback=status_callback: self.check_input_folder_step3(
                name, src, callback
            )
        )

        if progress_callback is not None:
            dicom_src_work.signal.sig_loop.connect(
                lambda i, maximum=dicom_src_work.get_files_len() + 1: progress_callback(
                    i, maximum
                )
            )
        QThreadPool.globalInstance().start(dicom_src_work)

    def check_input_folder_step3(
        self,
        data_input: DataInputList,
        dicom_src_work: DicomSearchWorker,
        status_callback: callable = None,
    ):
        """
        Updates SWANe UI at the end of the DICOM files scan Worker execution for a subject.

        Parameters
        ----------
        data_input : DataInputList
            The series folder name to check.
        dicom_src_work : DicomSearchWorker
            The DICOM Search Worker.
        status_callback: callable.
            The function to notify update to UI. Default is None

        Returns
        -------
        None.

        """

        subjects_list = dicom_src_work.tree.get_subject_list()

        if len(subjects_list) == 0:
            status_callback(
                data_input, SubjectRet.DataInputWarningNoDicom, dicom_src_work
            )
            return

        if len(subjects_list) > 1:
            status_callback(
                data_input, SubjectRet.DataInputWarningMultiSubj, dicom_src_work
            )
            return

        studies_list = dicom_src_work.tree.get_studies_list(subjects_list[0])

        if len(studies_list) != 1:
            status_callback(
                data_input, SubjectRet.DataInputWarningMultiStudy, dicom_src_work
            )
            return

        series_list = dicom_src_work.tree.get_series_list(
            subjects_list[0], studies_list[0]
        )

        if len(series_list) != 1:
            status_callback(
                data_input, SubjectRet.DataInputWarningMultiSeries, dicom_src_work
            )
            return

        series = dicom_src_work.tree.get_series(
            subjects_list[0], studies_list[0], series_list[0]
        )

        self.input_state_list[data_input].loaded = True
        self.input_state_list[data_input].volumes = series.volumes
        status_callback(data_input, SubjectRet.DataInputValid, dicom_src_work)

    def dicom_import_to_folder(
        self,
        data_input: DataInputList,
        copy_list: list,
        vols: int,
        mod: str,
        force_modality: bool,
        progress_callback: callable = None,
    ) -> SubjectRet:
        """
        Copies the files inside the selected folder in the input list into the folder specified by data_input var.

        Parameters
        ----------
        data_input : DataInputList
            The name of the series to which couple the selected file.

        Returns
        -------
        None.

        """
        # series already loaded
        if self.input_state_list[data_input].loaded:
            return SubjectRet.DataInputNonEmpty
        # number of volumes check
        if data_input.value.max_volumes != -1 and vols > data_input.value.max_volumes:
            return SubjectRet.DataImportErrorVolumesMax
        if vols < data_input.value.min_volumes:
            return SubjectRet.DataImportErrorVolumesMin

        # modality check
        if (
            not data_input.value.is_image_modality(ImageModality.from_string(mod))
            and not force_modality
        ):
            return SubjectRet.DataImportErrorModality

        dest_path = os.path.join(self.dicom_folder(), str(data_input))

        try:
            for thisFile in copy_list:
                if not os.path.isfile(thisFile):
                    continue

                shutil.copy(thisFile, dest_path)
                if progress_callback is not None:
                    progress_callback(1)

            return SubjectRet.DataImportCompleted
        except:
            return SubjectRet.DataImportErrorCopy

    def create_config(self, dependency_manager: DependencyManager):
        """
        Generate the subject configuration reading its config file.

        Parameters
        ----------
        dependency_manager: DependencyManager
            the application dependency manager

        Returns
        -------
        None.

        """
        self.config = ConfigManager(self.folder)
        self.config.check_dependencies(dependency_manager)

    def dicom_folder(self, data_input: DataInputList = None) -> str:
        """
        Get a dicom folder path

        Parameters
        ----------
        data_input: DataInputList, optional
            if specified, return the relative dicom subfolder

        Returns
        -------
            The requested dicom folder path

        """

        if type(data_input) is not DataInputList:
            return os.path.join(
                self.folder, self.global_config.get_default_dicom_folder()
            )
        else:
            return os.path.join(
                self.folder,
                self.global_config.get_default_dicom_folder(),
                str(data_input),
            )

    def dicom_folder_count(self, data_input: DataInputList) -> int:
        """
        Counts files in a dicom folder

        Parameters
        ----------
        data_input: DataInputList
            the data input dicom folder to scan

        Returns
        -------
            The file count as an int

        """
        try:
            dicom_path = self.dicom_folder(data_input)
            count = len(
                [
                    entry
                    for entry in os.listdir(dicom_path)
                    if os.path.isfile(os.path.join(dicom_path, entry))
                ]
            )
            return count
        except:
            return 0

    def check_subject_folder(self, subject_folder: str) -> SubjectRet:
        """
        Check if a path is a valid subject folder

        Parameters
        ----------
        subject_folder: str
            the path to check

        Returns
        -------
            A SubjectRet code

        """
        if not os.path.exists(subject_folder):
            return SubjectRet.FolderNotFound

        if " " in subject_folder:
            return SubjectRet.PathBlankSpaces

        if not os.path.abspath(subject_folder).startswith(
            os.path.abspath(self.global_config.get_main_working_directory() + os.sep)
        ):
            return SubjectRet.FolderOutsideMain

        if not self.check_subject_subtree(subject_folder):
            return SubjectRet.InvalidFolderTree

        return SubjectRet.ValidFolder

    def clear_import_folder(self, data_input: DataInputList) -> bool:
        """
        Clears the subject series folder.

        Parameters
        ----------
        data_input : DataInputList
            The series folder name to clear.

        Returns
        -------
        False if exception raised, True otherwise.

        """
        try:
            src_path = self.dicom_folder(data_input)

            shutil.rmtree(src_path, ignore_errors=True)
            os.makedirs(src_path, exist_ok=True)

            # Reset the workflows related to the deleted DICOM images
            src_path = os.path.join(
                self.folder,
                self.name + strings.WF_DIR_SUFFIX,
                data_input.value.workflow_name,
            )
            shutil.rmtree(src_path, ignore_errors=True)
            self.input_state_list[data_input].loaded = False
            self.input_state_list[data_input].volumes = 0
            return True
        except:
            return False

    def check_subject_subtree(self, subject_folder: str) -> bool:
        """
        Check if a directory is a valid subject folder

        Parameters
        ----------
        subject_folder : str
            The directory path to check.

        Returns
        -------
        bool
            True if the directory is a valid subject folder, otherwise False.

        """

        for data_input in DataInputList:
            if not os.path.exists(
                os.path.join(
                    subject_folder,
                    self.global_config.get_default_dicom_folder(),
                    str(data_input),
                )
            ):
                return False

        return True

    def fix_subject_folder_subtree(self, subject_folder: str):
        """
        Update an existing folder with the subject subfolder structure.

        Parameters
        ----------
        subject_folder : str
            The directory path to update into a subject folder.

        Returns
        -------
        None.

        """

        for data_input in DataInputList:
            if not os.path.exists(
                os.path.join(
                    subject_folder,
                    self.global_config.get_default_dicom_folder(),
                    str(data_input),
                )
            ):
                os.makedirs(
                    os.path.join(
                        subject_folder,
                        self.global_config.get_default_dicom_folder(),
                        str(data_input),
                    ),
                    exist_ok=True,
                )

    def create_new_subject_dir(self, subject_name: str) -> SubjectRet:
        """
        Create a new subject folder and subfolders.

        Parameters
        ----------
        subject_name : str
            The subject folder name.

        Returns
        -------
        True if no Exception raised.

        """
        invalid_chars = r"\/:*?<>|"

        if subject_name is None or subject_name == "":
            return SubjectRet.FolderNotFound
        elif (
            any(char in invalid_chars for char in subject_name)
            or subject_name.isspace()
            or " " in subject_name
        ):
            return SubjectRet.PathBlankSpaces
        elif os.path.exists(
            os.path.join(self.global_config.get_main_working_directory(), subject_name)
        ):
            return SubjectRet.FolderAlreadyExists
        else:
            try:
                base_folder = os.path.abspath(
                    os.path.join(
                        self.global_config.get_main_working_directory(), subject_name
                    )
                )
                dicom_folder = os.path.join(
                    base_folder, self.global_config.get_default_dicom_folder()
                )
                for data_input in DataInputList:
                    os.makedirs(
                        os.path.join(dicom_folder, str(data_input)), exist_ok=True
                    )
                return self.load(base_folder)
            except:
                return SubjectRet.FolderNotFound

    def can_generate_workflow(self) -> bool:
        """
        Check if requisites for workflow generation are met

        Returns
        -------
        True if workflow can be generated
        """
        return (
            self.input_state_list.is_ref_loaded()
            and self.dependency_manager.is_fsl()
            and self.dependency_manager.is_dcm2niix()
        )

    def graph_dir(self) -> str:
        """
        Returns
        -------
        The path of graph directory
        """
        return os.path.join(self.folder, Subject.GRAPH_DIR_NAME)

    def graph_file(self, long_name: str):
        """
        Parameters
        ----------
        long_name: str
            The workflow complete name
        Returns
        -------
        The complete path of graph file with the specified name
        """
        graph_name = long_name.lower().replace(" ", "_")
        return os.path.join(
            self.graph_dir(),
            Subject.GRAPH_FILE_PREFIX + graph_name + "." + Subject.GRAPH_FILE_EXT,
        )

    def result_dir(self) -> str:
        """
        Returns
        -------
        The subject results directory path
        """
        return os.path.join(self.folder, MainWorkflow.Result_DIR)

    def scene_path(self) -> str:
        """
        Returns
        -------
        The slicer scene file path
        """
        return os.path.join(
            self.result_dir(), "scene." + self.global_config.get_slicer_scene_ext()
        )

    def generate_workflow(self, generate_graphs: bool = True) -> SubjectRet:
        """
        Generates and populates the Main Workflow.
        Generates the graphviz analysis graphs on a new thread.

        Parameters
        ----------
        generate_graphs: bool.
            If True, svg graphics of workflows are generated. Default is True.

        Returns
        -------
        SubjectRet corresponding to success or failure

        """

        if not self.can_generate_workflow():
            return SubjectRet.GenWfMissingRequisites

        # Main Workflow generation
        if self.workflow is None:

            # Node List population
            try:
                self.workflow = MainWorkflow(
                    name=self.name + strings.WF_DIR_SUFFIX,
                    base_dir=self.folder,
                    global_config=self.global_config,
                    subject_config=self.config,
                    dependency_manager=self.dependency_manager,
                    subject_input_state_list=self.input_state_list,
                )
            except:
                traceback.print_exc()
                # TODO: generiamo un file crash nella cartella log?
                return SubjectRet.GenWfError

        graph_dir = self.graph_dir()
        shutil.rmtree(graph_dir, ignore_errors=True)
        os.mkdir(graph_dir)

        node_list = self.workflow.get_node_array()

        # Graphviz analysis graphs drawing
        if generate_graphs:
            for node in node_list.keys():
                if len(node_list[node].node_list.keys()) > 0:
                    if self.dependency_manager.is_graphviz():
                        thread = Thread(
                            target=self.workflow.get_node(node).write_graph,
                            kwargs={
                                "graph2use": self.GRAPH_TYPE,
                                "format": Subject.GRAPH_FILE_EXT,
                                "dotfilename": os.path.join(
                                    self.graph_file(node_list[node].long_name)
                                ),
                            },
                        )
                        thread.start()

        return SubjectRet.GenWfCompleted

    def is_workflow_process_alive(self) -> bool:
        """
        Checks if a workflow is in execution.

        Returns
        -------
        bool
            True if the workflow is executing, elsewise False.

        """

        try:
            if self.workflow_process is None:
                return False
            return self.workflow_process.is_alive()
        except AttributeError:
            return False

    def workflow_dir(self) -> str:
        """
        Returns
        -------
        The workflow directory path
        """
        return os.path.join(self.folder, self.name + strings.WF_DIR_SUFFIX)

    def workflow_dir_exists(self) -> bool:
        """
        Returns
        -------
        True if a previous workflow execution is found
        """
        return os.path.exists(self.workflow_dir())

    def delete_workflow_dir(self):
        """
        Delete any previous workflow execution
        """
        shutil.rmtree(self.workflow_dir(), ignore_errors=True)
        self.result_dir()

    def delete_result_dir(self):
        """
        Delete any previous result
        """
        shutil.rmtree(self.result_dir(), ignore_errors=True)

    def freesurfer_dir(self) -> str:
        """
        Returns
        -------
        The freesurfer subject directory path
        """
        return os.path.join(self.folder, FS_DIR)

    def freesurfer_dir_exists(self) -> bool:
        """
        Returns
        -------
        True if freesurfer subject directory exists
        """
        return os.path.exists(self.freesurfer_dir())

    def delete_freesurfer_dir(self):
        """
        Delete any previous freesurfer run
        """
        shutil.rmtree(self.freesurfer_dir(), ignore_errors=True)

    def start_workflow(
        self,
        resume: bool = None,
        resume_freesurfer: bool = None,
        update_node_callback: callable = None,
    ) -> SubjectRet:
        """
        Start the workflow execution in a subprocess

        Parameters
        ----------
        resume: bool, optional
            If True resume previous run, if False delete them, if None and previous run is found, stops. Default is None
        resume_freesurfer: bool, optional
            If True resume previous fs run , if False delete them, if None and previous fs run is found, stops. Default is None
        update_node_callback: callable, optional
            The method to notify workflow update

        Returns
        -------
        A SubjectRet code

        """
        # Already executing workflow
        if self.is_workflow_process_alive():
            return SubjectRet.ExecWfStatusError
        # Checks for a previous workflow execution
        if self.workflow_dir_exists():
            if resume is None:
                return SubjectRet.ExecWfResume
            elif not resume:
                self.delete_workflow_dir()
                self.delete_result_dir()

        # Checks for a previous workflow FreeSurfer execution
        if self.config.get_workflow_freesurfer_pref() and self.freesurfer_dir_exists():
            if resume_freesurfer is None:
                return SubjectRet.ExecWfResumeFreesurfer
            elif not resume_freesurfer:
                self.delete_freesurfer_dir()

        queue = Queue(maxsize=500)

        # Generates a Monitor Worker to receive workflows notifications
        self.workflow_monitor_work = WorkflowMonitorWorker(queue)
        if update_node_callback is not None:
            self.workflow_monitor_work.signal.log_msg.connect(update_node_callback)
        QThreadPool.globalInstance().start(self.workflow_monitor_work)

        # Starts the workflow on a new process
        self.workflow_process = WorkflowProcess(self.name, self.workflow, queue)
        self.workflow_process.start()
        return SubjectRet.ExecWfStarted

    def stop_workflow(self) -> SubjectRet:
        """
        Stop a running workflow execution

        Returns
        -------
        A SubjectRet code

        """
        if not self.is_workflow_process_alive():
            return SubjectRet.ExecWfStatusError
        # Workflow killing
        self.workflow_process.stop_event.set()

    def reset_workflow(self, force: bool = False) -> bool:
        """
        Set the workflow var to None.
        Resets the UI.
        Works only if the worklow is not in execution or if force var is True.

        Parameters
        ----------
        force : bool, optional
            Force the usage of this function during workflow execution. The default is False.

        Returns
        -------
        True if workflow was generated

        """

        if self.workflow is None:
            return False
        if not force and self.is_workflow_process_alive():
            return False

        self.workflow = None
        return True

    def generate_scene(self, progress_callback: callable = None):
        """
        Exports the workflow results into 3D Slicer using a new thread.

        Returns
        -------
        None.

        """

        slicer_thread = SlicerExportWorker(
            self.global_config.get_slicer_path(),
            self.result_dir(),
            self.global_config.get_slicer_scene_ext(),
            self.config,
        )
        if progress_callback is not None:
            slicer_thread.signal.export.connect(progress_callback)
        QThreadPool.globalInstance().start(slicer_thread)



import os
import sys
from swane import strings
import subprocess
from swane.utils.platform_and_tools_utils import is_command_available, is_linux, is_mac


FSL_CONFLICT_PATH = "fsl/bin"
FREESURFER_CONFIG_FILE = "SetUpFreeSurfer.sh"
SHELL_PROFILE = {
    "sh": [".profile"],
    "bash": [".bash_profile", ".profile", ".bashrc"],
    "dash": [".bash_profile", ".profile"],
    "zsh": [".zprofile", ".zshrc"],
    "csh": [".cshrc"],
    "tcsh": [".tcshrc"],
}

FIX_LINE = r"""PATH=$(echo "$PATH" | sed -e "s/:$( echo "$FSL_DIR" | sed 's/\//\\\//g')\/bin//")"""
APP_EXEC_COMMAND = "python3 -m " + __name__.split(".")[0]


def check_config_file(config_file: str):
    try:
        with open(config_file) as file:
            file_content = file.read()
        return FREESURFER_CONFIG_FILE in file_content
    except:
        return False


def get_config_file():
    # Try to identify user shell configuration file with fsl/freesurfer setup
    try:
        shell = os.path.basename(os.environ.get("SHELL", "sh")).lower()
        home_dir = os.path.expanduser("~")

        if shell not in SHELL_PROFILE:
            shell = "sh"

        candidates = [os.path.join(home_dir, p) for p in SHELL_PROFILE[shell]]
        for candidate in candidates:
            if check_config_file(candidate):
                return candidate

        return strings.generic_shell_file

    except KeyError:
        return strings.generic_shell_file


def runtime_fix():
    cmd = FIX_LINE + ";" + APP_EXEC_COMMAND
    subprocess.run(cmd, shell=True)


def config_file_fix(config_file: str):
    with open(config_file, "a") as file_object:
        file_object.write("\n" + FIX_LINE)


def copy_fix_to_clipboard():
    # Linux shell copy command
    if is_linux():
        subprocess.run("xclip -selection c", shell=True, text=True, input=FIX_LINE)
    # MacOS shell copy command
    if is_mac():
        subprocess.run("pbcopy", shell=True, text=True, input=FIX_LINE)


def fsl_conflict_check() -> bool:
    """
        Handle freesurfer<=7.3.2 overwriting system python executable if fsl>=6.0.6 is installed
        In that case propose a fix

    Returns
    -------
        True if system Python is unaffected, False if it was hidden

    """

    if FSL_CONFLICT_PATH not in sys.executable:
        return True

    # This function uses fsl built-in qt library to show a warning: ignore IDE import error!
    from PyQt5.QtWidgets import QApplication, QMessageBox

    app = QApplication([])
    app.setApplicationDisplayName(strings.APPNAME)

    config_file = get_config_file()
    error_string = strings.fsl_python_error % config_file

    msg_box = QMessageBox()
    msg_box.setText(error_string)
    msg_box.setInformativeText(FIX_LINE)
    msg_box.setIcon(QMessageBox.Warning)
    msg_box.setStandardButtons(QMessageBox.Yes | QMessageBox.Retry | QMessageBox.Cancel)
    msg_box.button(QMessageBox.Yes).setText(strings.fsl_python_error_fix)
    msg_box.button(QMessageBox.Retry).setText(strings.fsl_python_error_restart)

    if is_command_available("xclip") or is_mac():
        msg_box.button(QMessageBox.Cancel).setText(strings.fsl_python_error_exit)
        msg_box.setDefaultButton(QMessageBox.Cancel)
    ret = msg_box.exec()

    if ret == QMessageBox.Retry:
        runtime_fix()
    elif ret == QMessageBox.Yes:
        config_file_fix(config_file)
        runtime_fix()
    elif ret == QMessageBox.Cancel and (is_command_available("xclip") or is_mac()):
        copy_fix_to_clipboard()

    return False



from nipype.interfaces.freesurfer import ReconAll
from nipype.interfaces.fsl import ImageMaths, ImageStats, ApplyMask, BinaryMaths

from nipype import Node, Merge

from swane.nipype_pipeline.engine.CustomWorkflow import CustomWorkflow
from swane.nipype_pipeline.nodes.TTest import TTest
from swane.nipype_pipeline.nodes.SumMultiVols import SumMultiVols
from swane.nipype_pipeline.nodes.utils import getn

from nipype.interfaces.utility import IdentityInterface

lh_labels = [
    17,  # Left-Hippocampus                    220 216 20  0
    18,  # Left-Amygdala                       103 255 255 0
    1000,  # ctx-lh-unknown                      25  5   25  0
    1001,  # ctx-lh-bankssts                     25  100 40  0
    1002,  # ctx-lh-caudalanteriorcingulate      125 100 160 0
    1003,  # ctx-lh-caudalmiddlefrontal          100 25  0   0
    1004,  # ctx-lh-corpuscallosum               120 70  50  0
    1005,  # ctx-lh-cuneus                       220 20  100 0
    1006,  # ctx-lh-entorhinal                   220 20  10  0
    1007,  # ctx-lh-fusiform                     180 220 140 0
    1008,  # ctx-lh-inferiorparietal             220 60  220 0
    1009,  # ctx-lh-inferiortemporal             180 40  120 0
    1010,  # ctx-lh-isthmuscingulate             140 20  140 0
    1011,  # ctx-lh-lateraloccipital             20  30  140 0
    1012,  # ctx-lh-lateralorbitofrontal         35  75  50  0
    1013,  # ctx-lh-lingual                      225 140 140 0
    1014,  # ctx-lh-medialorbitofrontal          200 35  75  0
    1015,  # ctx-lh-middletemporal               160 100 50  0
    1016,  # ctx-lh-parahippocampal              20  220 60  0
    1017,  # ctx-lh-paracentral                  60  220 60  0
    1018,  # ctx-lh-parsopercularis              220 180 140 0
    1019,  # ctx-lh-parsorbitalis                20  100 50  0
    1020,  # ctx-lh-parstriangularis             220 60  20  0
    1021,  # ctx-lh-pericalcarine                120 100 60  0
    1022,  # ctx-lh-postcentral                  220 20  20  0
    1023,  # ctx-lh-posteriorcingulate           220 180 220 0
    1024,  # ctx-lh-precentral                   60  20  220 0
    1025,  # ctx-lh-precuneus                    160 140 180 0
    1026,  # ctx-lh-rostralanteriorcingulate     80  20  140 0
    1027,  # ctx-lh-rostralmiddlefrontal         75  50  125 0
    1028,  # ctx-lh-superiorfrontal              20  220 160 0
    1029,  # ctx-lh-superiorparietal             20  180 140 0
    1030,  # ctx-lh-superiortemporal             140 220 220 0
    1031,  # ctx-lh-supramarginal                80  160 20  0
    1032,  # ctx-lh-frontalpole                  100 0   100 0
    1033,  # ctx-lh-temporalpole                 70  70  70  0
    1034,  # ctx-lh-transversetemporal           150 150 200 0
    1035,  # ctx-lh-insula                       255 192 32  0
]

labels_rh = [
    53,  # Right - Hippocampus                 220 216 20  0
    54,  # Right - Amygdala                    103 255 255 0
    2000,  # ctx-rh-unknown                      25  5   25  0
    2001,  # ctx-rh-bankssts                     25  100 40  0
    2002,  # ctx-rh-caudalanteriorcingulate      125 100 160 0
    2003,  # ctx-rh-caudalmiddlefrontal          100 25  0   0
    2004,  # ctx-rh-corpuscallosum               120 70  50  0
    2005,  # ctx-rh-cuneus                       220 20  100 0
    2006,  # ctx-rh-entorhinal                   220 20  10  0
    2007,  # ctx-rh-fusiform                     180 220 140 0
    2008,  # ctx-rh-inferiorparietal             220 60  220 0
    2009,  # ctx-rh-inferiortemporal             180 40  120 0
    2010,  # ctx-rh-isthmuscingulate             140 20  140 0
    2011,  # ctx-rh-lateraloccipital             20  30  140 0
    2012,  # ctx-rh-lateralorbitofrontal         35  75  50  0
    2013,  # ctx-rh-lingual                      225 140 140 0
    2014,  # ctx-rh-medialorbitofrontal          200 35  75  0
    2015,  # ctx-rh-middletemporal               160 100 50  0
    2016,  # ctx-rh-parahippocampal              20  220 60  0
    2017,  # ctx-rh-paracentral                  60  220 60  0
    2018,  # ctx-rh-parsopercularis              220 180 140 0
    2019,  # ctx-rh-parsorbitalis                20  100 50  0
    2020,  # ctx-rh-parstriangularis             220 60  20  0
    2021,  # ctx-rh-pericalcarine                120 100 60  0
    2022,  # ctx-rh-postcentral                  220 20  20  0
    2023,  # ctx-rh-posteriorcingulate           220 180 220 0
    2024,  # ctx-rh-precentral                   60  20  220 0
    2025,  # ctx-rh-precuneus                    160 140 180 0
    2026,  # ctx-rh-rostralanteriorcingulate     80  20  140 0
    2027,  # ctx-rh-rostralmiddlefrontal         75  50  125 0
    2028,  # ctx-rh-superiorfrontal              20  220 160 0
    2029,  # ctx-rh-superiorparietal             20  180 140 0
    2030,  # ctx-rh-superiortemporal             140 220 220 0
    2031,  # ctx-rh-supramarginal                80  160 20  0
    2032,  # ctx-rh-frontalpole                  100 0   100 0
    2033,  # ctx-rh-temporalpole                 70  70  70  0
    2034,  # ctx-rh-transversetemporal           150 150 200 0
    2035,  # ctx-rh-insula                       255 192 32  0
]


def get_symmetric(index):
    if index >= 1000:
        return index + 1000
    if index == 17:
        return 53
    if index == 18:
        return 54
    return -1


def freesurfer_asymmetry_index_workflow(
    name: str, base_dir: str = "/"
) -> CustomWorkflow:
    """
    Freesurfer cortical reconstruction, white matter ROI, basal ganglia and thalami ROI.
    If needed, segmentation of the hippocampal substructures and the nuclei of the amygdala.

    Parameters
    ----------
    name : str
        The workflow name.
    base_dir : path, optional
        The base directory path relative to parent workflow. The default is "/".

    Input Node Fields
    ----------
    in_file : path
        The input file to use for AI calculation.
    seg_file : path
        Freesurfer segmentation file.

    Returns
    -------
    workflow : CustomWorkflow
        The Freesurfer workflow.

    Output Node Fields
    ----------
    asymmetry_index_file : string
        The asymmetry index map.

    """

    workflow = CustomWorkflow(name=name, base_dir=base_dir)

    # Input Node
    inputnode = Node(
        IdentityInterface(fields=["in_file", "seg_file"]), name="inputnode"
    )

    # Output Node
    outputnode = Node(
        IdentityInterface(
            fields=["asymmetry_t", "asymmetry_p", "asymmetry_z", "asymmetry_ai"]
        ),
        name="outputnode",
    )

    merge_node_t = Node(Merge(len(lh_labels) * 2), name="merge_node_t")
    merge_node_p = Node(Merge(len(lh_labels) * 2), name="merge_node_p")
    merge_node_z = Node(Merge(len(lh_labels) * 2), name="merge_node_z")
    merge_node_ai = Node(Merge(len(lh_labels) * 2), name="merge_node_ai")
    index = 1

    for lh_label in lh_labels:

        rh_label = get_symmetric(lh_label)
        if rh_label == -1:
            continue

        lh_mask = Node(ImageMaths(), name="lh_mask_%d" % lh_label)
        lh_mask.inputs.op_string = "-thr %d -uthr %d -bin" % (lh_label, lh_label)
        workflow.add_nodes([lh_mask])
        workflow.connect(inputnode, "seg_file", lh_mask, "in_file")

        lh_apply_mask = Node(ApplyMask(), name="lh_applymask_%d" % lh_label)
        workflow.connect(inputnode, "in_file", lh_apply_mask, "in_file")
        workflow.connect(lh_mask, "out_file", lh_apply_mask, "mask_file")

        lh_stats = Node(ImageStats(), name="lh_stats_%d" % lh_label)
        lh_stats.inputs.op_string = "-M -S -V"
        workflow.connect(lh_apply_mask, "out_file", lh_stats, "in_file")

        rh_mask = Node(ImageMaths(), name="rh_mask_%d" % rh_label)
        rh_mask.inputs.op_string = "-thr %d -uthr %d -bin" % (rh_label, rh_label)
        workflow.add_nodes([rh_mask])
        workflow.connect(inputnode, "seg_file", rh_mask, "in_file")

        rh_mask_inv = Node(ImageMaths(), name="rh_mask_inv_%d" % rh_label)
        rh_mask_inv.inputs.op_string = "-mul -1"
        workflow.connect(rh_mask, "out_file", rh_mask_inv, "in_file")

        rh_apply_mask = Node(ApplyMask(), name="rh_applymask_%d" % rh_label)
        workflow.connect(inputnode, "in_file", rh_apply_mask, "in_file")
        workflow.connect(rh_mask, "out_file", rh_apply_mask, "mask_file")

        rh_stats = Node(ImageStats(), name="rh_stats_%d" % rh_label)
        rh_stats.inputs.op_string = "-M -S -V"
        workflow.connect(rh_apply_mask, "out_file", rh_stats, "in_file")

        t_test = Node(TTest(), name="ttest_%d" % lh_label)
        workflow.connect(lh_stats, "out_stat", t_test, "stats_lh")
        workflow.connect(rh_stats, "out_stat", t_test, "stats_rh")

        lh_ai_sum = Node(BinaryMaths(), name="lh_ai_sum_%d" % lh_label)
        lh_ai_sum.inputs.operation = "add"
        workflow.connect(inputnode, "in_file", lh_ai_sum, "in_file")
        workflow.connect(rh_stats, ("out_stat", getn, 0), lh_ai_sum, "operand_value")

        lh_ai_sub = Node(BinaryMaths(), name="lh_ai_sub_%d" % lh_label)
        lh_ai_sub.inputs.operation = "sub"
        workflow.connect(inputnode, "in_file", lh_ai_sub, "in_file")
        workflow.connect(rh_stats, ("out_stat", getn, 0), lh_ai_sub, "operand_value")

        lh_ai = Node(BinaryMaths(), name="lh_ai_%d" % lh_label)
        lh_ai.inputs.operation = "div"
        workflow.connect(lh_ai_sub, "out_file", lh_ai, "in_file")
        workflow.connect(lh_ai_sum, "out_file", lh_ai, "operand_file")

        lh_ai_mask = Node(ApplyMask(), name="lh_ai_applymask_%d" % lh_label)
        workflow.connect(lh_ai, "out_file", lh_ai_mask, "in_file")
        workflow.connect(lh_mask, "out_file", lh_ai_mask, "mask_file")

        rh_ai_sum = Node(BinaryMaths(), name="rh_ai_sum_%d" % rh_label)
        rh_ai_sum.inputs.operation = "add"
        workflow.connect(inputnode, "in_file", rh_ai_sum, "in_file")
        workflow.connect(rh_stats, ("out_stat", getn, 0), rh_ai_sum, "operand_value")

        rh_ai_sub = Node(BinaryMaths(), name="rh_ai_sub_%d" % rh_label)
        rh_ai_sub.inputs.operation = "sub"
        workflow.connect(inputnode, "in_file", rh_ai_sub, "in_file")
        workflow.connect(rh_stats, ("out_stat", getn, 0), rh_ai_sub, "operand_value")

        rh_ai = Node(BinaryMaths(), name="rh_ai_%d" % rh_label)
        rh_ai.inputs.operation = "div"
        workflow.connect(rh_ai_sub, "out_file", rh_ai, "in_file")
        workflow.connect(rh_ai_sum, "out_file", rh_ai, "operand_file")

        rh_ai_mask = Node(ApplyMask(), name="rh_ai_applymask_%d" % rh_label)
        workflow.connect(rh_ai, "out_file", rh_ai_mask, "in_file")
        workflow.connect(rh_mask, "out_file", rh_ai_mask, "mask_file")

        lh_value_t = Node(BinaryMaths(), name="lh_value_t_%d" % lh_label)
        lh_value_t.inputs.operation = "mul"
        workflow.connect(lh_mask, "out_file", lh_value_t, "in_file")
        workflow.connect(t_test, "stat_t", lh_value_t, "operand_value")

        rh_value_t = Node(BinaryMaths(), name="rh_value_t_%d" % lh_label)
        rh_value_t.inputs.operation = "mul"
        workflow.connect(rh_mask_inv, "out_file", rh_value_t, "in_file")
        workflow.connect(t_test, "stat_t", rh_value_t, "operand_value")

        lh_value_p = Node(BinaryMaths(), name="lh_value_p_%d" % lh_label)
        lh_value_p.inputs.operation = "mul"
        workflow.connect(lh_mask, "out_file", lh_value_p, "in_file")
        workflow.connect(t_test, "stat_p", lh_value_p, "operand_value")

        rh_value_p = Node(BinaryMaths(), name="rh_value_p_%d" % lh_label)
        rh_value_p.inputs.operation = "mul"
        workflow.connect(rh_mask_inv, "out_file", rh_value_p, "in_file")
        workflow.connect(t_test, "stat_p", rh_value_p, "operand_value")

        def get_z_op_string(stat_list):
            return "-sub %f -div %f -mas" % (stat_list[0], stat_list[1])

        lh_z = Node(ImageMaths(), name="lh_z_%d" % lh_label)
        workflow.connect(inputnode, "in_file", lh_z, "in_file")
        workflow.connect(rh_stats, ("out_stat", get_z_op_string), lh_z, "op_string")
        workflow.connect(lh_mask, "out_file", lh_z, "in_file2")

        rh_z = Node(ImageMaths(), name="rh_z_%d" % rh_label)
        workflow.connect(inputnode, "in_file", rh_z, "in_file")
        workflow.connect(lh_stats, ("out_stat", get_z_op_string), rh_z, "op_string")
        workflow.connect(rh_mask, "out_file", rh_z, "in_file2")

        workflow.connect(lh_value_t, "out_file", merge_node_t, "in%d" % index)
        workflow.connect(lh_value_p, "out_file", merge_node_p, "in%d" % index)
        workflow.connect(lh_z, "out_file", merge_node_z, "in%d" % index)
        workflow.connect(lh_ai_mask, "out_file", merge_node_ai, "in%d" % index)
        index += 1
        workflow.connect(rh_value_t, "out_file", merge_node_t, "in%d" % index)
        workflow.connect(lh_value_p, "out_file", merge_node_p, "in%d" % index)
        workflow.connect(rh_z, "out_file", merge_node_z, "in%d" % index)
        workflow.connect(rh_ai_mask, "out_file", merge_node_ai, "in%d" % index)
        index += 1

    sum_masks_t = Node(SumMultiVols(), name="sum_masks_t")
    workflow.connect(merge_node_t, "out", sum_masks_t, "vol_files")

    sum_masks_p = Node(SumMultiVols(), name="sum_masks_p")
    workflow.connect(merge_node_p, "out", sum_masks_p, "vol_files")

    sum_masks_z = Node(SumMultiVols(), name="sum_masks_z")
    workflow.connect(merge_node_z, "out", sum_masks_z, "vol_files")

    sum_masks_ai = Node(SumMultiVols(), name="sum_masks_ai")
    workflow.connect(merge_node_ai, "out", sum_masks_ai, "vol_files")

    workflow.connect(sum_masks_t, "out_file", outputnode, "asymmetry_t")
    workflow.connect(sum_masks_p, "out_file", outputnode, "asymmetry_p")
    workflow.connect(sum_masks_z, "out_file", outputnode, "asymmetry_z")
    workflow.connect(sum_masks_ai, "out_file", outputnode, "asymmetry_ai")

    return workflow



import pydicom
import os
from PySide6.QtCore import Signal, QObject, QRunnable
from swane.nipype_pipeline.MainWorkflow import DEBUG
from swane.utils.DicomTree import DicomTree
from dicom_sequence_classifier import extract_metadata, load_dicom_file, classify_dicom


class DicomSearchSignal(QObject):
    sig_loop = Signal(int)
    sig_finish = Signal(object)


class DicomSearchWorker(QRunnable):

    def __init__(self, dicom_dir: str, classify: bool = False):
        """
        Thread class to scan a dicom folder and return dicom files ordered in subjects, exams and series

        Parameters
        ----------
        dicom_dir: str
            The dicom folder to scan
        classify: bool
            Try to classify dicom images in series. Default is False
        """
        super(DicomSearchWorker, self).__init__()
        if os.path.exists(os.path.abspath(dicom_dir)):
            self.dicom_dir = os.path.abspath(dicom_dir)
            self.unsorted_list = []
        self.signal = DicomSearchSignal()
        self.tree = DicomTree(dicom_dir)
        self.error_message = []
        self.classify = classify
        # self.dicom_tree = {}
        # self.series_positions = {}
        # self.multi_frame_series = {}

    @staticmethod
    def clean_text(string: str) -> str:
        """
        Remove forbidden characters from a string

        Parameters
        ----------
        string: str
            The string to clean.

        Returns
            The cleaned string in lower case.
        -------

        """
        # clean and standardize text descriptions, which makes searching files easier
        forbidden_symbols = [
            "*",
            ".",
            ",",
            '"',
            "\\",
            "/",
            "|",
            "[",
            "]",
            ":",
            ";",
            " ",
        ]
        for symbol in forbidden_symbols:
            # replace everything with an underscore
            string = string.replace(symbol, "_")
        return string.lower()

    def load_dir(self):
        """
        Generates the list of file to be scanned.
        """
        if (
            self.dicom_dir is None
            or self.dicom_dir == ""
            or not os.path.exists(self.dicom_dir)
        ):
            return
        self.unsorted_list = []
        for root, dirs, files in os.walk(self.dicom_dir):
            for file in files:
                self.unsorted_list.append(os.path.join(root, file))

    def get_files_len(self):
        """
        The number of file to be scanned
        """
        try:
            return len(self.unsorted_list)
        except:
            return 0

    def run(self):
        try:
            if len(self.unsorted_list) == 0:
                self.load_dir()

            skip = False

            for dicom_loc in self.unsorted_list:
                self.signal.sig_loop.emit(1)

                if skip:
                    continue

                # read the file
                if not os.path.exists(dicom_loc):
                    continue
                ds = pydicom.dcmread(dicom_loc, force=True)

                subject_id = ds.get("PatientID", "na")
                if subject_id == "na":
                    continue

                series_number = ds.get("SeriesNumber", "NA")
                study_instance_uid = ds.get("StudyInstanceUID", "NA")

                # in GE la maggior parte delle ricostruzioni sono DERIVED\SECONDARY
                if (
                    hasattr(ds, "ImageType")
                    and "DERIVED" in ds.ImageType
                    and "SECONDARY" in ds.ImageType
                    and "ASL" not in ds.ImageType
                ):
                    if ds.ImageType not in self.error_message:
                        self.error_message.append(ds.ImageType)
                    continue
                # in GE e SIEMENS l'immagine anatomica di ASL è ORIGINAL\PRIMARY\ASL
                if (
                    hasattr(ds, "ImageType")
                    and "ORIGINAL" in ds.ImageType
                    and "PRIMARY" in ds.ImageType
                    and "ASL" in ds.ImageType
                ):
                    if ds.ImageType not in self.error_message:
                        self.error_message.append(ds.ImageType)
                    continue
                # in Philips e Siemens le ricostruzioni sono PROJECTION IMAGE
                if hasattr(ds, "ImageType") and "PROJECTION IMAGE" in ds.ImageType:
                    if ds.ImageType not in self.error_message:
                        self.error_message.append(ds.ImageType)
                    continue

                self.tree.add_subject(subject_id, str(ds.PatientName))
                self.tree.add_study(subject_id, study_instance_uid)
                dicom_series = self.tree.add_series(
                    subject_id, study_instance_uid, series_number
                )

                multi_frame_series = False
                if "NumberOfFrames" in ds and int(ds.NumberOfFrames) > 1:
                    multi_frame_series = True

                dicom_series.add_dicom_loc(
                    dicom_loc, multi_frame_series, ds.get("SliceLocation"), ds
                )
                dicom_series.modality = ds.Modality
                if dicom_series.description == "Not named":
                    if hasattr(ds, "SeriesDescription"):
                        dicom_series.description = ds.SeriesDescription
                    else:
                        dicom_series.description = (
                            DicomSearchWorker.find_series_description(
                                dicom_series.dicom_locs
                            )
                        )

                # TODO: calcolare multiframe alla fine

                if self.classify and dicom_series.classification == "Not classified":
                    dicom_series.classification = (
                        DicomSearchWorker.find_series_classification(ds)
                    )

            for subject in self.tree.dicom_subjects:
                for study in self.tree.dicom_subjects[subject].studies:
                    for series in self.tree.dicom_subjects[subject].studies[study]:
                        self.tree.dicom_subjects[subject].studies[study][
                            series
                        ].refine_frame_number()

            self.signal.sig_loop.emit(1)
            self.signal.sig_finish.emit(self)
        except:
            self.signal.sig_finish.emit(self)

    @staticmethod
    def find_series_description(image_list: list[str]) -> str:
        """
        Extract the description of the dicom series searching among all the series images.
        The description is equal to:
        - the SeriesDescription tag, if any in one of the image list
        - otherwise, None (unnamed_series)

        Parameters
        ----------
        image_list: list[str]
            The dicom file list to check

        Returns
        -------
        str
            The dicom series description

        """

        for image in image_list:
            ds = pydicom.dcmread(image, force=True)

            if hasattr(ds, "SeriesDescription"):
                return ds.SeriesDescription
        return "Unnamed series"

    @staticmethod
    def find_series_classification(ds) -> str:
        """
        Analyses the dicom using dicom_sequence_classifier to attempt an automatic dicom series classification.

        Parameters
        ----------
        ds:
            The dicom dataset to check

        Returns
        -------
        str
            The dicom series classification

        """

        meta = extract_metadata(ds)
        classification = classify_dicom(meta)
        if classification != "NOT MR":
            return classification

        return "Unknown"



import os
from swane.utils.DataInputList import DataInputList, FMRI_NUM
from swane import __version__
from multiprocessing import cpu_count
from nipype.utils.profiler import get_system_total_memory_gb
from math import ceil

from swane import strings
from swane.config.PreferenceEntry import PreferenceEntry
from swane.config.config_enums import (
    InputTypes,
    WORKFLOW_TYPES,
    SLICER_EXTENSIONS,
    CORE_LIMIT,
    VEIN_DETECTION_MODE,
    BLOCK_DESIGN,
    SLICE_TIMING,
    GlobalPrefCategoryList,
    BETWEEN_MOD_FLIRT_COST,
)

try:
    base_dir = os.path.abspath(os.path.join(os.environ["FSLDIR"], "data/xtract_data"))
    human_path = os.path.join(base_dir, "Human")
    HUMAN_path = os.path.join(base_dir, "HUMAN")

    if os.path.exists(human_path):
        # For FSL Version < 6.0.7
        XTRACT_DATA_DIR = human_path
    elif os.path.exists(HUMAN_path):
        # For FSL Version >= 6.0.7
        XTRACT_DATA_DIR = HUMAN_path
    else:
        XTRACT_DATA_DIR = ""
except:
    XTRACT_DATA_DIR = ""
DEFAULT_N_SAMPLES = 5000

TRACTS = {
    "af": ["Arcuate Fasciculus", "true", 0],
    "ar": ["Acoustic Radiation", "false", 0],
    "atr": ["Anterior Thalamic Radiation", "false", 0],
    "cbd": ["Cingulum subsection : Dorsal", "false", 0],
    "cbp": ["Cingulum subsection : Peri-genual", "false", 0],
    "cbt": ["Cingulum subsection : Temporal", "false", 0],
    "cst": ["Corticospinal Tract", "true", 0],
    "fa": ["Frontal Aslant", "false", 0],
    "fma": ["Forceps Major", "false", 0],
    "fmi": ["Forceps Minor", "false", 0],
    "fx": ["Fornix", "false", 0],
    "ilf": ["Inferior Longitudinal Fasciculus", "false", 0],
    "ifo": ["Inferior Fronto-Occipital Fasciculus", "false", 0],
    "mcp": ["Middle Cerebellar Peduncle", "false", 0],
    "mdlf": ["Middle Longitudinal Fasciculus", "false", 0],
    "or": ["Optic Radiation", "true", 0],
    "str": ["Superior Thalamic Radiation", "false", 0],
    "ac": ["Anterior Commissure", "false", 0],
    "uf": ["Uncinate Fasciculus", "false", 0],
    "vof": ["Vertical Occipital Fasciculus", "false", 0],
}
structure_file = os.path.join(XTRACT_DATA_DIR, "structureList")
if os.path.exists(structure_file):
    with open(structure_file, "r") as file:
        for line in file.readlines():
            split = line.split(" ")
            tract_name = split[0][:-2]
            if tract_name in tuple(TRACTS.keys()):
                try:
                    TRACTS[tract_name][2] = int(float(split[1]) * 1000)
                except:
                    TRACTS[tract_name][2] = DEFAULT_N_SAMPLES

for k in list(TRACTS.keys()):
    if TRACTS[k][2] == 0:
        del TRACTS[k]

# WORKFLOW_TYPES = ["Structural Workflow", "Morpho-Functional Workflow"]
# SLICER_EXTENSIONS = ["mrb", "mrml"]

# WORKFLOWS PREFERENCE LIST
WF_PREFERENCES = {}

category = DataInputList.T13D
WF_PREFERENCES[category] = {}
WF_PREFERENCES[category]["wf_type"] = PreferenceEntry(
    input_type=InputTypes.ENUM,
    hidden=True,
    label="Default workflow",
    value_enum=WORKFLOW_TYPES,
    default=WORKFLOW_TYPES.STRUCTURAL,
)
WF_PREFERENCES[category]["bet_bias_correction"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    label="Bias reduction for skull removal",
    tooltip="Increase time with better results",
    default="false",
)
WF_PREFERENCES[category]["bet_thr"] = PreferenceEntry(
    input_type=InputTypes.FLOAT,
    label="Threshold value for skull removal",
    default=0.3,
    tooltip="Accepted values from 0 to 1, higher values are considered equal 1",
    range=[0, 1],
)
WF_PREFERENCES[category]["freesurfer"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    label="FreeSurfer analysis",
    default="true",
    dependency="is_freesurfer",
    dependency_fail_tooltip="Freesurfer not detected",
)
WF_PREFERENCES[category]["hippo_amyg_labels"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    label="FreeSurfer hippocampal and amygdala subfields",
    default="false",
    dependency="is_freesurfer_matlab",
    dependency_fail_tooltip="Matlab Runtime not detected",
    pref_requirement={DataInputList.T13D: [("freesurfer", True)]},
    pref_requirement_fail_tooltip="Requires Freesurfer analysis",
)
WF_PREFERENCES[category]["flat1"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    label="FLAT1 analysis",
    default="true",
    input_requirement=[DataInputList.FLAIR3D],
    input_requirement_fail_tooltip="Requires both 3D T1w and 3D Flair",
)

category = DataInputList.FLAIR3D
WF_PREFERENCES[category] = {}
WF_PREFERENCES[category]["bet_bias_correction"] = WF_PREFERENCES[DataInputList.T13D][
    "bet_bias_correction"
]
WF_PREFERENCES[category]["bet_thr"] = PreferenceEntry(
    input_type=InputTypes.FLOAT,
    label="Threshold value for skull removal",
    default=0.5,
    tooltip="Accepted values from 0 to 1, higher values are considered equal 1",
    range=[0, 1],
)

category = DataInputList.MDC
WF_PREFERENCES[category] = {}
WF_PREFERENCES[category]["bet_bias_correction"] = WF_PREFERENCES[DataInputList.T13D][
    "bet_bias_correction"
]
WF_PREFERENCES[category]["bet_thr"] = PreferenceEntry(
    input_type=InputTypes.FLOAT,
    label="Threshold value for skull removal",
    default=0.5,
    tooltip="Accepted values from 0 to 1, higher values are considered equal 1",
    range=[0, 1],
)

category = DataInputList.VENOUS
WF_PREFERENCES[category] = {}
WF_PREFERENCES[category]["bet_thr"] = PreferenceEntry(
    input_type=InputTypes.FLOAT,
    label="Threshold value for skull removal",
    default=0.4,
    tooltip="Accepted values from 0 to 1, higher values are considered equal 1",
    range=[0, 1],
)
WF_PREFERENCES[category]["vein_detection_mode"] = PreferenceEntry(
    input_type=InputTypes.ENUM,
    label="Venous volume detection mode",
    value_enum=VEIN_DETECTION_MODE,
    default=VEIN_DETECTION_MODE.SD,
)
WF_PREFERENCES[category]["vein_segment_threshold"] = PreferenceEntry(
    input_type=InputTypes.FLOAT,
    label="Threshold (%) for 3DSlicer Vein Segment",
    default=97.5,
    range=[0.1, 100],
    decimals=1,
)

category = DataInputList.ASL
WF_PREFERENCES[category] = {}
WF_PREFERENCES[category]["cost_func"] = PreferenceEntry(
    input_type=InputTypes.ENUM,
    label="FLIRT between modalities cost function",
    value_enum=BETWEEN_MOD_FLIRT_COST,
    default=BETWEEN_MOD_FLIRT_COST.NORMALIZED_MUTUAL_INFORMATION,
)
WF_PREFERENCES[category]["ai"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    label="Asymmetry Index map for ASL",
    default="true",
)
WF_PREFERENCES[category]["ai_threshold"] = PreferenceEntry(
    input_type=InputTypes.INT,
    label="Threshold for Asymmetry Index map outliers removal",
    tooltip="100 for no thresholding, suggested 80-90",
    default=85,
    range=[0, 100],
    pref_requirement={DataInputList.ASL: [("ai", True)]},
    pref_requirement_fail_tooltip="Requires ASL Asymmetry Index",
)

category = DataInputList.PET
WF_PREFERENCES[category] = {}
WF_PREFERENCES[category]["cost_func"] = PreferenceEntry(
    input_type=InputTypes.ENUM,
    label="FLIRT between modalities cost function",
    value_enum=BETWEEN_MOD_FLIRT_COST,
    default=BETWEEN_MOD_FLIRT_COST.MULTUAL_INFORMATION,
)
WF_PREFERENCES[category]["ai"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    label="Asymmetry Index map for PET",
    default="true",
)
WF_PREFERENCES[category]["ai_threshold"] = PreferenceEntry(
    input_type=InputTypes.INT,
    label="Threshold for Asymmetry Index map outliers removal",
    tooltip="100 for no thresholding, suggested 80-90",
    default=85,
    range=[0, 100],
    pref_requirement={DataInputList.PET: [("ai", True)]},
    pref_requirement_fail_tooltip="Requires PET Asymmetry Index",
)
category = DataInputList.DTI
WF_PREFERENCES[category] = {}
WF_PREFERENCES[category]["old_eddy_correct"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    label="Use older but faster fsl eddy_correct",
    default="false",
)
WF_PREFERENCES[category]["tractography"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    label="DTI tractography",
    default="true",
)
WF_PREFERENCES[category]["tractography_threshold"] = PreferenceEntry(
    input_type=InputTypes.FLOAT,
    label="Threshold for 3DSlicer DTI Tract",
    tooltip="This value is multiplied by the tract waytotal for threshold calculation",
    default=0.0035,
    range=[0.0001, 1],
    decimals=4,
)
WF_PREFERENCES[category]["track_procs"] = PreferenceEntry(
    input_type=InputTypes.INT,
    label="Parallel processes for each side tractography",
    default=5,
    range=[1, 10],
    pref_requirement={DataInputList.DTI: [("tractography", True)]},
    pref_requirement_fail_tooltip="Tractography disabled",
)

for tract in TRACTS.keys():
    WF_PREFERENCES[category][tract] = PreferenceEntry(
        input_type=InputTypes.BOOLEAN,
        label=TRACTS[tract][0],
        default=TRACTS[tract][1],
        pref_requirement={DataInputList.DTI: [("tractography", True)]},
        pref_requirement_fail_tooltip="Tractography disabled",
    )

for x in range(FMRI_NUM):
    category = DataInputList["FMRI" + "_%s" % x]
    WF_PREFERENCES[category] = {}
    WF_PREFERENCES[category]["task_a_name"] = PreferenceEntry(
        input_type=InputTypes.TEXT,
        label="Task A name",
        default="Task_A",
    )
    WF_PREFERENCES[category]["block_design"] = PreferenceEntry(
        input_type=InputTypes.ENUM,
        label="Block design",
        value_enum=BLOCK_DESIGN,
        default=BLOCK_DESIGN.RARA,
    )
    WF_PREFERENCES[category]["task_b_name"] = PreferenceEntry(
        input_type=InputTypes.TEXT,
        label="Task B name",
        default="Task_B",
        pref_requirement={
            DataInputList["FMRI" + "_%s" % x]: [("block_design", BLOCK_DESIGN.RARB)]
        },
        pref_requirement_fail_tooltip="Requires rArBrArB... block design",
    )
    WF_PREFERENCES[category]["task_duration"] = PreferenceEntry(
        input_type=InputTypes.INT,
        label="Tasks duration (sec)",
        default=30,
        range=[1, 500],
    )
    WF_PREFERENCES[category]["rest_duration"] = PreferenceEntry(
        input_type=InputTypes.INT,
        label="Rest duration (sec)",
        default=30,
        range=[0, 500],
    )
    WF_PREFERENCES[category]["tr"] = PreferenceEntry(
        input_type=InputTypes.FLOAT,
        label="Repetition Time (TR)",
        tooltip="Set -1 for automatic detection",
        default="-1.0",
        range=[-1, 1000],
    )
    WF_PREFERENCES[category]["n_vols"] = PreferenceEntry(
        input_type=InputTypes.INT,
        label="Number of volumes",
        tooltip="Set -1 for automatic detection",
        default="-1",
        range=[-1, 1000],
    )
    WF_PREFERENCES[category]["slice_timing"] = PreferenceEntry(
        input_type=InputTypes.ENUM,
        label="Slice timing",
        value_enum=SLICE_TIMING,
        default=SLICE_TIMING.UNKNOWN,
    )
    WF_PREFERENCES[category]["del_start_vols"] = PreferenceEntry(
        input_type=InputTypes.INT,
        label="Delete start volumes",
        default=0,
        range=[0, 500],
    )
    WF_PREFERENCES[category]["del_end_vols"] = PreferenceEntry(
        input_type=InputTypes.INT,
        label="Delete end volumes",
        default=0,
        range=[0, 500],
    )

GLOBAL_PREFERENCES = {}

category = GlobalPrefCategoryList.MAIN
GLOBAL_PREFERENCES[category] = {}
GLOBAL_PREFERENCES[category]["main_working_directory"] = PreferenceEntry(
    input_type=InputTypes.DIRECTORY,
    label="Main working directory",
    box_text="Select the main working directory",
    default="",
    restart=True,
)
GLOBAL_PREFERENCES[category]["subjects_prefix"] = PreferenceEntry(
    input_type=InputTypes.TEXT,
    hidden=True,
    default="subj_",
)
GLOBAL_PREFERENCES[category]["slicer_path"] = PreferenceEntry(
    input_type=InputTypes.FILE,
    label="3D Slicer path",
    box_text="Select 3D Slicer executable",
    default="",
    restart=True,
    validate_on_change=True,
)
GLOBAL_PREFERENCES[category]["slicer_version"] = PreferenceEntry(
    input_type=InputTypes.TEXT,
    hidden=True,
    default="",
)
GLOBAL_PREFERENCES[category]["last_pid"] = PreferenceEntry(
    input_type=InputTypes.INT,
    hidden=True,
    default=-1,
)
GLOBAL_PREFERENCES[category]["last_swane_version"] = PreferenceEntry(
    input_type=InputTypes.TEXT,
    hidden=True,
    default=__version__,
)
GLOBAL_PREFERENCES[category]["force_pref_reset"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    hidden=True,
    default="false",
)
GLOBAL_PREFERENCES[category]["slicer_scene_ext"] = PreferenceEntry(
    input_type=InputTypes.ENUM,
    hidden=True,
    value_enum=SLICER_EXTENSIONS,
    default=SLICER_EXTENSIONS.MRB,
)
GLOBAL_PREFERENCES[category]["default_dicom_folder"] = PreferenceEntry(
    input_type=InputTypes.TEXT,
    hidden=True,
    default="dicom",
)
GLOBAL_PREFERENCES[category]["default_wf_type"] = PreferenceEntry(
    input_type=InputTypes.ENUM,
    label="Default workflow",
    value_enum=WORKFLOW_TYPES,
    default=WORKFLOW_TYPES.STRUCTURAL,
)
GLOBAL_PREFERENCES[category]["shutdown"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    hidden=True,
    label=strings.menu_shutdown_pref,
    default="false",
    default_at_startup="true",
)
GLOBAL_PREFERENCES[category]["auto_import"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    label=strings.menu_auto_import,
    default="true",
)
category = GlobalPrefCategoryList.PERFORMANCE
GLOBAL_PREFERENCES[category] = {}
GLOBAL_PREFERENCES[category]["max_subj"] = PreferenceEntry(
    input_type=InputTypes.INT,
    label="Patient tab limit",
    default=1,
    range=[0, 5],
)
try:
    suggested_max_cpu = max(
        ceil(min(cpu_count() / 2, get_system_total_memory_gb() / 3)), 1
    )
except:
    suggested_max_cpu = 1
GLOBAL_PREFERENCES[category]["max_subj_cu"] = PreferenceEntry(
    input_type=InputTypes.INT,
    label="CPU core limit per subject",
    tooltip="To use all CPU cores set value equal to -1",
    default=str(suggested_max_cpu),
    range=[-1, 30],
)
GLOBAL_PREFERENCES[category]["resource_monitor"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    label="Enable resource monitor",
    default="false",
)
GLOBAL_PREFERENCES[category]["multicore_node_limit"] = PreferenceEntry(
    input_type=InputTypes.ENUM,
    label="CPU management for multi-core steps",
    value_enum=CORE_LIMIT,
    default=CORE_LIMIT.SOFT_CAP,
    informative_text={
        CORE_LIMIT.NO_LIMIT: "Multi-core steps ignore the subject CPU core limit, using all available resources",
        CORE_LIMIT.SOFT_CAP: "Multi-core steps use up to twice the subject CPU core limit",
        CORE_LIMIT.HARD_CAP: "Multi-core steps strictly respect the subject CPU core limit",
    },
)
GLOBAL_PREFERENCES[category]["cuda"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    label="Enable CUDA for GPUable commands",
    tooltip="NVIDIA GPU-based computation",
    default="false",
    dependency="is_cuda",
    dependency_fail_tooltip="GPU does not support CUDA",
)
GLOBAL_PREFERENCES[category]["max_subj_gpu"] = PreferenceEntry(
    input_type=InputTypes.INT,
    label="GPU process limit per subject",
    tooltip="The limit should be equal or lesser than the number of physical GPU",
    default=1,
    range=[1, 5],
    pref_requirement={GlobalPrefCategoryList.PERFORMANCE: [("cuda", True)]},
    pref_requirement_fail_tooltip="Requires CUDA",
)
category = GlobalPrefCategoryList.OPTIONAL_SERIES
GLOBAL_PREFERENCES[category] = {}
for data_input in DataInputList:
    if data_input.value.optional:
        GLOBAL_PREFERENCES[category][data_input.value.name] = PreferenceEntry(
            input_type=InputTypes.BOOLEAN,
            label=data_input.value.label,
            default="false",
            restart=True,
        )

category = GlobalPrefCategoryList.MAIL_SETTINGS
GLOBAL_PREFERENCES[category] = {}
GLOBAL_PREFERENCES[category]["enabled"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    label="Enabled",
    tooltip="Toggle on/off the mail report sending service. If enabled, a mail report will be sent to your email at each workflow completion",
    default="false",
)
GLOBAL_PREFERENCES[category]["address"] = PreferenceEntry(
    input_type=InputTypes.TEXT,
    label="Address",
    tooltip="The POP3/IMAP address of your mail host",
    default="",
)
GLOBAL_PREFERENCES[category]["port"] = PreferenceEntry(
    input_type=InputTypes.INT,
    label="Port",
    tooltip="The port indicated in your mail host documentation",
    default=0,
    range=[1, 1000],
)
GLOBAL_PREFERENCES[category]["username"] = PreferenceEntry(
    input_type=InputTypes.TEXT, label="Username", tooltip="Your email", default=""
)
GLOBAL_PREFERENCES[category]["password"] = PreferenceEntry(
    input_type=InputTypes.PASSWORD,
    label="Password",
    tooltip="Your password email",
    default="",
)
GLOBAL_PREFERENCES[category]["use_ssl"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    label="Use SSL",
    tooltip="True if your mail host requires the SSL security protocol",
    default="true",
)
GLOBAL_PREFERENCES[category]["use_tls"] = PreferenceEntry(
    input_type=InputTypes.BOOLEAN,
    label="Use TLS",
    tooltip="True if your mail host requires the TLS security protocol. TLS is not used if SSL is active",
    default="false",
)

DEFAULT_WF = {}
DEFAULT_WF[WORKFLOW_TYPES.STRUCTURAL] = {
    DataInputList.T13D: {
        "hippo_amyg_labels": "false",
        "flat1": "false",
    },
    DataInputList.DTI: {
        "tractography": "true",
    },
    DataInputList.ASL: {"ai": "false"},
    DataInputList.PET: {"ai": "false"},
}
DEFAULT_WF[WORKFLOW_TYPES.FUNCTIONAL] = {
    DataInputList.T13D: {
        "hippo_amyg_labels": "true",
        "flat1": "true",
    },
    DataInputList.DTI: {
        "tractography": "false",
    },
    DataInputList.ASL: {"ai": "true"},
    DataInputList.PET: {"ai": "true"},
}



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype.interfaces.fsl.epi import Eddy, EddyInputSpec
from nipype.utils.gpu_count import gpu_count
from shutil import which


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.epi.Eddy)  -*-
class CustomEddy(Eddy):
    """
    Custom implementation of Eddy subclass to support use_gpu input.

    """

    _cmd = "eddy_openmp" if which("eddy_openmp") else "eddy_cpu"

    def _use_cuda(self):
        if self.inputs.use_cuda and gpu_count() > 0:
            self.inputs.num_threads = 1
            # eddy_cuda usually link to eddy_cudaX.X but some versions miss the symlink
            # anyway in newer fsl versions eddy automatically use cuda on cuda-capable systems
            self._cmd = "eddy_cuda" if which("eddy_cuda") else "eddy"
        else:
            # older fsl versions has cuda_openmp, newer versions has eddy_cpu
            self._cmd = "eddy_openmp" if which("eddy_openmp") else "eddy_cpu"

    def _num_threads_update(self):
        if self.inputs.use_cuda:
            self.inputs.num_threads = 1
        super()._num_threads_update()



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype.interfaces.base import (
    TraitedSpec,
    File,
    BaseInterfaceInputSpec,
    BaseInterface,
)
import os


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.fsl.base.FSLCommandInputSpec)  -*-
class GenEddyFilesInputSpec(BaseInterfaceInputSpec):
    bval = File(exists=True, mandatory=True, desc="the bval file")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class GenEddyFilesOutputSpec(TraitedSpec):
    index = File(desc="Volume index file for Eddy")
    acqp = File(desc="Acquisition parameters file for Eddy")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.fsl.base.FSLCommand)  -*-
class GenEddyFiles(BaseInterface):
    """
    Generates index and acqp files for eddy. Does not support topup

    """

    input_spec = GenEddyFilesInputSpec
    output_spec = GenEddyFilesOutputSpec

    def _run_interface(self, runtime):
        bval_file = open(self.inputs.bval, "r")
        bvals = bval_file.read().split()
        bval_file.close()
        self.index_file = os.path.abspath("index")
        with open(self.index_file, "w") as file:
            for val in bvals:
                file.write("1 ")
        self.acqp_file = os.path.abspath("acqp")
        with open(self.acqp_file, "w") as file:
            file.write("0 1 0 0.05")

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["index"] = self.index_file
        outputs["acqp"] = self.acqp_file
        return outputs



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype.interfaces.fsl.base import FSLCommand, FSLCommandInputSpec
from nipype.interfaces.base import traits, TraitedSpec, File, isdefined


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.fsl.base.FSLCommandInputSpec)  -*-
class FslNVolsInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True, mandatory=True, argstr="%s", position="1", desc="the input image"
    )
    force_value = traits.Int(mandatory=False, desc="value forced by user")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class FslNVolsOutputSpec(TraitedSpec):
    nvols = traits.Int(desc="Number of EPI runs")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.FSLCommand)  -*-
class FslNVols(FSLCommand):
    """
    Reads the num. of volumes from a 4d NIFTI file.

    """

    _cmd = "fslnvols"
    input_spec = FslNVolsInputSpec
    output_spec = FslNVolsOutputSpec

    def aggregate_outputs(self, runtime=None, needed_outputs=None):
        outputs = self._outputs()

        # se l'utente ha inserito un valore forzo quello invece del risultato della lettura automatica
        if isdefined(self.inputs.force_value) and self.inputs.force_value != -1:
            outputs.nvols = self.inputs.force_value
            return outputs

        info = runtime.stdout
        try:
            outputs.nvols = int(info)
        except ValueError:
            outputs.nvols = 0

        return outputs



import uuid
from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
import base64


class CryptographyManager:
    """
    Criptography manager service

    Attributes
    ----------
    cryptography_key : str
        The key to use for the encryption/decryption processes.

    """

    password = str(uuid.getnode()).encode()
    salt = b"\x84mt\xec\xcc\xd1\n\xe5\xe8\x86`\xce\xe3\x9aa\xcd"
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=390000,
    )
    cryptography_key = base64.urlsafe_b64encode(kdf.derive(password))

    @staticmethod
    def encrypt(string: str) -> str:
        fernet = Fernet(CryptographyManager.cryptography_key)
        encMessage = fernet.encrypt(string.encode())
        return encMessage.decode()

    @staticmethod
    def decrypt(string: str) -> str:
        fernet = Fernet(CryptographyManager.cryptography_key)
        decMessage = fernet.decrypt(string.encode()).decode()
        return str(decMessage)



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype.pipeline.engine import Workflow
from nipype import Node, logging
from nipype.interfaces.utility import IdentityInterface
from nipype.interfaces.io import DataSink
from swane.nipype_pipeline.engine.NodeListEntry import NodeListEntry
from swane import strings

logger = logging.getLogger("nipype.workflow")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.pipeline.engine.Workflow)  -*-
class CustomWorkflow(Workflow):
    """
    Custom implementation of Workflow class with utility funcs.

    """

    @staticmethod
    def format_node_name(node):
        """
        Returns the explicit name of a Node.

        """

        default_node_name = None
        if (
            hasattr(node, "interface")
            and type(node.interface).__name__ in strings.node_names
        ):
            default_node_name = strings.node_names[type(node.interface).__name__]
        if hasattr(node, "long_name"):
            formatted_name = node.long_name
            if "%s" in node.long_name and default_node_name is not None:
                formatted_name = node.long_name % default_node_name
        elif default_node_name is not None:
            formatted_name = default_node_name
        else:
            formatted_name = node.name

        formatted_name = formatted_name[0].upper() + formatted_name[1:]
        return formatted_name

    def get_node_array(self) -> dict:
        """
        Returns a List of NodeListEntry objects for the Nodes in a Workflow.

        """

        from networkx import topological_sort

        outlist = {}
        for node in topological_sort(self._graph):
            if hasattr(node, "interface") and isinstance(
                node.interface, IdentityInterface
            ):
                continue

            outlist[node.name] = NodeListEntry()
            outlist[node.name].long_name = self.format_node_name(node)
            if isinstance(node, CustomWorkflow):
                outlist[node.name].node_list = node.get_node_array()
        return outlist

    def sink_result(
        self,
        save_path: str,
        result_node: str,
        result_name: str,
        sub_folder: str,
        regexp_substitutions: list[tuple[str, str]] = None,
    ):
        """
        Creates a sink_result Node to save the output files of a Workflow.

        """

        if isinstance(result_node, str):
            result_node = self.get_node(result_node)

        data_sink = Node(
            DataSink(),
            name="SaveResults_"
            + result_node.name
            + "_"
            + result_name.replace(".", "_"),
        )
        data_sink.long_name = "%s: " + result_name
        data_sink.inputs.base_directory = save_path

        if regexp_substitutions is not None:
            data_sink.inputs.regexp_substitutions = regexp_substitutions

        self.connect(result_node, result_name, data_sink, sub_folder)

    def _get_dot(
        self, prefix=None, hierarchy=None, colored=False, simple_form=True, level=0
    ):
        """
        Custom implementation of _get_dot Nipype func to support the long_name Node attribute.

        """

        import networkx as nx

        if prefix is None:
            prefix = "  "
        if hierarchy is None:
            hierarchy = []
        colorset = [
            "#FFFFC8",  # Y
            "#0000FF",
            "#B4B4FF",
            "#E6E6FF",  # B
            "#FF0000",
            "#FFB4B4",
            "#FFE6E6",  # R
            "#00A300",
            "#B4FFB4",
            "#E6FFE6",  # G
            "#0000FF",
            "#B4B4FF",
        ]  # loop B
        if level > len(colorset) - 2:
            level = 3  # Loop back to blue

        dotlist = ['%slabel="%s";' % (prefix, self.name)]
        for node in nx.topological_sort(self._graph):
            fullname = ".".join(hierarchy + [node.fullname])
            nodename = fullname.replace(".", "_")
            if not isinstance(node, Workflow):
                node_class_name = self.format_node_name(node)
                if hasattr(node, "iterables") and node.iterables:
                    dotlist.append(
                        (
                            '%s[label="%s", shape=box3d,'
                            "style=filled, color=black, colorscheme"
                            "=greys7 fillcolor=2];"
                        )
                        % (nodename, node_class_name)
                    )
                else:
                    if colored:
                        dotlist.append(
                            ('%s[label="%s", style=filled,' ' fillcolor="%s"];')
                            % (nodename, node_class_name, colorset[level])
                        )
                    else:
                        dotlist.append(
                            ('%s[label="%s"];') % (nodename, node_class_name)
                        )

        for node in nx.topological_sort(self._graph):
            if isinstance(node, Workflow):
                fullname = ".".join(hierarchy + [node.fullname])
                nodename = fullname.replace(".", "_")
                dotlist.append("subgraph cluster_%s {" % nodename)
                if colored:
                    dotlist.append(
                        prefix + prefix + 'edge [color="%s"];' % (colorset[level + 1])
                    )
                    dotlist.append(prefix + prefix + "style=filled;")
                    dotlist.append(
                        prefix + prefix + 'fillcolor="%s";' % (colorset[level + 2])
                    )
                dotlist.append(
                    node._get_dot(
                        prefix=prefix + prefix,
                        hierarchy=hierarchy + [self.name],
                        colored=colored,
                        simple_form=simple_form,
                        level=level + 3,
                    )
                )
                dotlist.append("}")
            else:
                for subnode in self._graph.successors(node):
                    if node._hierarchy != subnode._hierarchy:
                        continue
                    if not isinstance(subnode, Workflow):
                        nodefullname = ".".join(hierarchy + [node.fullname])
                        subnodefullname = ".".join(hierarchy + [subnode.fullname])
                        nodename = nodefullname.replace(".", "_")
                        subnodename = subnodefullname.replace(".", "_")
                        for _ in self._graph.get_edge_data(node, subnode)["connect"]:
                            dotlist.append("%s -> %s;" % (nodename, subnodename))
                        logger.debug("connection: %s", dotlist[-1])
        # add between workflow connections
        for u, v, d in self._graph.edges(data=True):
            uname = ".".join(hierarchy + [u.fullname])
            vname = ".".join(hierarchy + [v.fullname])
            for src, dest in d["connect"]:
                uname1 = uname
                vname1 = vname
                if isinstance(src, tuple):
                    srcname = src[0]
                else:
                    srcname = src
                if "." in srcname:
                    uname1 += "." + ".".join(srcname.split(".")[:-1])
                if "." in dest and "@" not in dest:
                    if not isinstance(v, Workflow):
                        if "datasink" not in str(v._interface.__class__).lower():
                            vname1 += "." + ".".join(dest.split(".")[:-1])
                    else:
                        vname1 += "." + ".".join(dest.split(".")[:-1])
                if uname1.split(".")[:-1] != vname1.split(".")[:-1]:
                    dotlist.append(
                        "%s -> %s;"
                        % (uname1.replace(".", "_"), vname1.replace(".", "_"))
                    )
                    logger.debug("cross connection: %s", dotlist[-1])
        return ("\n" + prefix).join(dotlist)



from PySide6.QtWidgets import (
    QWidget,
    QLabel,
    QHBoxLayout,
    QTreeWidgetItem,
    QSpacerItem,
    QTreeWidget,
)
from PySide6.QtSvgWidgets import QSvgWidget
from swane import strings


class CustomTreeWidgetItem(QTreeWidgetItem):
    """
    Custom implementation of PySide QTreeWidgetItem to define the Workflows Tree items.

    """

    def __init__(self, parent, tree: QTreeWidget, text: str, art: str = None):
        super(CustomTreeWidgetItem, self).__init__(parent)

        self.widget = QWidget()
        self.widget.setLayout(QHBoxLayout())
        self.artLabel = QSvgWidget()
        self.artLabel.setFixedWidth(26)
        self.artLabel.setFixedHeight(26)
        self.set_art(art)
        self.textLabel = QLabel(text)
        self.resize_text_label()

        self.widget.layout().addWidget(self.artLabel)
        self.widget.layout().addWidget(self.textLabel)
        self.widget.layout().addSpacerItem(QSpacerItem(25, 0))

        tree.setItemWidget(self, 0, self.widget)

        self.completed = False
        self.art = None

    def setToolTip(self, column, toolTip):
        """
        Reimplement base method to add info character .

        """
        if toolTip == "" and self.get_text().endswith(strings.INFOCHAR):
            self.set_text(self.get_text()[:-2])
        elif toolTip != "" and not self.get_text().endswith(strings.INFOCHAR):
            self.set_text(self.get_text() + " " + strings.INFOCHAR)
        super().setToolTip(column, toolTip)

    def set_text(self, text: str):
        """
        Set the tree item text.

        Parameters
        ----------
        text : str
            The item text.

        Returns
        -------
        None.

        """

        self.textLabel.setText(text)
        self.resize_text_label()

    def resize_text_label(self):
        """
        Resize the tree item label.
        Allow the horizontal scroll when expanding subsection.

        Returns
        -------
        None.

        """

        self.textLabel.setMinimumWidth(
            self.textLabel.fontMetrics().boundingRect(self.textLabel.text()).width()
            + 10
        )

    def get_text(self) -> str:
        """
        Get the tree item text from its label.

        Returns
        -------
        str
            The item text.

        """

        return self.textLabel.text()

    def set_art(self, art: str):
        """
        Set the icon of the tree item.

        Parameters
        ----------
        art : str
            The icon path of the tree item.

        Returns
        -------
        None.

        """

        self.art = art

        if art is not None:
            self.artLabel.load(art)



from PySide6.QtCore import QRunnable
import os
import subprocess


class SlicerViewerWorker(QRunnable):
    """
    Spawn a thread for 3D Slicer result export

    """

    PROGRESS_MSG_PREFIX = "SLICERLOADER: "
    END_MSG = "ENDLOADING"

    def __init__(self, slicer_path: str, scene_path: str):
        """
            Visualize the workflow results into 3D Slicer.

        Parameters
        -------
        slicer_path: str
           The slicer execution path
        scene_path: str
           The scene file path

        """
        super(SlicerViewerWorker, self).__init__()
        self.slicer_path: str = slicer_path
        self.scene_path: str = scene_path

    def run(self):
        cmd = (
            self.slicer_path
            + " --python-code 'slicer.util.loadScene(\""
            + self.scene_path
            + "\")'"
        )
        popen = subprocess.Popen(
            cmd,
            cwd=os.getcwd(),
            shell=True,
            stdout=subprocess.PIPE,
            universal_newlines=True,
        )



from PySide6.QtCore import Signal, QObject, QRunnable
from swane.nipype_pipeline.engine.WorkflowReport import WorkflowReport, WorkflowSignals
from multiprocessing import Queue


class LogReceiverSignal(QObject):
    log_msg = Signal(WorkflowReport)


class WorkflowMonitorWorker(QRunnable):
    """
    Create a thread waiting for nipype workflow reports using a multiprocessing queue.
    """

    def __init__(self, queue: Queue):
        super(WorkflowMonitorWorker, self).__init__()
        self.signal: LogReceiverSignal = LogReceiverSignal()
        self.queue: Queue = queue

    def run(self):
        while True:
            # get a unit of work
            wf_report = self.queue.get()
            # report
            self.signal.log_msg.emit(wf_report)
            # check for stop
            if wf_report.signal_type == WorkflowSignals.WORKFLOW_STOP:
                break



from nipype import Node, IdentityInterface
from nipype.interfaces.fsl import FLIRT, FNIRT, InvWarp, SwapDimensions

from swane.nipype_pipeline.engine.CustomWorkflow import CustomWorkflow


# TODO check base_dir = "./"
def nonlinear_reg_workflow(name: str, base_dir: str = "/") -> CustomWorkflow:
    """
    Transforms input images in a reference space through a nonlinear registration.
    For symmetric atlas, make a RL swapped to unswapped nonlinear registration.

    Parameters
    ----------
    name : str
        The workflow name.
    base_dir : path, optional
        The base directory path relative to parent workflow. The default is "/".

    Input Node Fields
    ----------
    atlas : path
        The standard atlas for the registration.
    in_file : path
        The input image for the registration.

    Returns
    -------
    workflow : CustomWorkflow
        The nonlinear registration workflow.

    Output Node Fields
    ----------
    fieldcoeff_file : path
        Nonlinear registration warp to atlas space.
    inverse_warp : path
        Nonlinear inverse registration warp from atlas space.
    out_matrix_file : path
        Linear registration matrix to atlas space.
    warped_file : path
        Input image transformed in atlas space.

    """

    workflow = CustomWorkflow(name=name, base_dir=base_dir)

    # Input Node
    inputnode = Node(IdentityInterface(fields=["atlas", "in_file"]), name="inputnode")

    # Output Node
    outputnode = Node(
        IdentityInterface(
            fields=["fieldcoeff_file", "inverse_warp", "out_matrix_file", "warped_file"]
        ),
        name="outputnode",
    )

    # NODE 1: Linear registration
    flirt = Node(FLIRT(), name="ref_2_%s_flirt" % name)
    flirt.long_name = "%s to atlas"
    flirt.inputs.searchr_x = [-90, 90]
    flirt.inputs.searchr_y = [-90, 90]
    flirt.inputs.searchr_z = [-90, 90]
    flirt.inputs.dof = 12
    # TODO consider switch to same-modality cost function
    flirt.inputs.cost = "corratio"
    flirt.inputs.out_matrix_file = "ref_2_%s.mat" % name
    workflow.add_nodes([flirt])
    workflow.connect(inputnode, "in_file", flirt, "in_file")
    workflow.connect(inputnode, "atlas", flirt, "reference")

    # NODE 2: Nonlinear registration
    fnirt = Node(FNIRT(), name="ref_2_%s_fnirt" % name)
    fnirt.long_name = "%s to atlas"
    fnirt.inputs.fieldcoeff_file = True
    workflow.connect(flirt, "out_matrix_file", fnirt, "affine_file")
    workflow.connect(inputnode, "in_file", fnirt, "in_file")
    workflow.connect(inputnode, "atlas", fnirt, "ref_file")

    # NODE 3: Inverse matrix
    invwarp = Node(InvWarp(), name="ref_2_%s_invwarp" % name)
    invwarp.long_name = "%s from atlas"
    workflow.connect(fnirt, "fieldcoeff_file", invwarp, "warp")
    workflow.connect(inputnode, "in_file", invwarp, "reference")

    workflow.connect(flirt, "out_matrix_file", outputnode, "out_matrix_file")
    workflow.connect(fnirt, "fieldcoeff_file", outputnode, "fieldcoeff_file")
    workflow.connect(fnirt, "warped_file", outputnode, "warped_file")
    workflow.connect(invwarp, "inverse_warp", outputnode, "inverse_warp")

    return workflow



from enum import Enum, auto


class WorkflowSignals(Enum):
    NODE_STARTED = auto()
    NODE_COMPLETED = auto()
    NODE_ERROR = auto()
    WORKFLOW_INSUFFICIENT_RESOURCES = auto()
    WORKFLOW_STOP = auto()
    INVALID_SIGNAL = auto()


class WorkflowReport:

    NODE_MSG_DIVIDER = "."

    def __init__(
        self,
        signal_type: WorkflowSignals = WorkflowSignals.NODE_STARTED,
        long_name: str = None,
        info: str = None,
    ):
        """

        Parameters
        ----------
        signal_type: int
            A signal type in SIGNAL_TYPES
        long_name: str
            Optional. The node longname.
        info: str
            Optional. An informative text. For future implementations.
        """

        if not isinstance(signal_type, WorkflowSignals):
            signal_type = WorkflowSignals.INVALID_SIGNAL
        self.signal_type = signal_type
        self.workflow_name = None
        self.node_name = None
        if long_name is not None:
            # Every longname is like "nipype_pt_x.workflow_name.node_name.message_type", we need second and third part
            split = long_name.split(WorkflowReport.NODE_MSG_DIVIDER)
            if len(split) == 3:
                self.workflow_name = split[1]
                long_name = split[2]
            self.node_name = long_name
        self.info = info



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype.interfaces.freesurfer import Label2Vol
from os.path import abspath


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.freesurfer.Label2Vol)  -*-
class CustomLabel2Vol(Label2Vol):
    """
    Custom implementation of Label2Vol Nipype Node to return the complete path of an output_file instead of only its name.

    """

    def _list_outputs(self):
        outputs = super(CustomLabel2Vol, self)._list_outputs()
        outputs["vol_label_file"] = abspath(outputs["vol_label_file"])
        return outputs



from swane.nipype_pipeline.engine.CustomWorkflow import CustomWorkflow
from swane.nipype_pipeline.nodes.CustomDcm2niix import CustomDcm2niix
from swane.nipype_pipeline.nodes.ForceOrient import ForceOrient
from swane.nipype_pipeline.nodes.CropFov import CropFov
from configparser import SectionProxy
from swane.utils.DataInputList import DataInputList
from nipype.interfaces.fsl import BET, RobustFOV
from nipype.interfaces.utility import IdentityInterface

from nipype import Node


def ref_workflow(
    name: str, dicom_dir: str, config: SectionProxy, base_dir: str = "/"
) -> CustomWorkflow:
    """
    T13D workflow to use as reference.

    Parameters
    ----------
    name : str
        The workflow name.
    dicom_dir : path
        The file path of the DICOM files.
    config: SectionProxy
        workflow settings.
    base_dir : path, optional
        The base directory path relative to parent workflow. The default is "/".

    Input Node Fields
    ----------
    -

    Returns
    -------
    workflow : CustomWorkflow
        The T13D reference workflow.

    Output Node Fields
    ----------
    ref : path
        T13D.
    ref_brain : path
        Betted T13D.
    ref_mask : path
        Brain mask from T13D bet command.

    """

    workflow = CustomWorkflow(name=name, base_dir=base_dir)

    # Output Node
    outputnode = Node(
        IdentityInterface(fields=["ref", "ref_brain", "ref_mask"]), name="outputnode"
    )

    # NODE 1: Conversion dicom -> nifti
    conversion = Node(CustomDcm2niix(), name="%s_conv" % name)
    conversion.inputs.source_dir = dicom_dir
    conversion.inputs.bids_format = False
    conversion.inputs.out_filename = "converted"
    conversion.inputs.name_conflicts = 1
    conversion.inputs.merge_imgs = 2

    # NODE 2: Orienting in radiological convention
    ref_reOrient = Node(ForceOrient(), name="%s_reOrient" % name)
    workflow.connect(conversion, "converted_files", ref_reOrient, "in_file")

    # NODE 3: Crop neck
    ref_robustfov = Node(RobustFOV(), name="%s_robustfov" % name)
    ref_robustfov.inputs.out_roi = "ref_robustfov.nii.gz"
    workflow.connect(ref_reOrient, "out_file", ref_robustfov, "in_file")

    # NODE 4: Crop FOV larger than 256mm for subsequent freesurfer
    ref_reScale = Node(CropFov(), name="%s_reScale" % name)
    ref_reScale.long_name = "Crop large FOV"
    ref_reScale.inputs.max_dim = 256
    ref_reScale.inputs.out_file = "ref.nii.gz"
    workflow.connect(ref_robustfov, "out_roi", ref_reScale, "in_file")

    # NODE 5: Scalp removal
    ref_BET = Node(BET(), name="%s_BET" % name)
    ref_BET.inputs.mask = True
    ref_BET.inputs.frac = config.getfloat_safe("bet_thr")
    if config.getboolean_safe("bet_bias_correction"):
        ref_BET.inputs.reduce_bias = True
    else:
        ref_BET.inputs.robust = True

    workflow.connect(ref_reScale, "out_file", ref_BET, "in_file")

    workflow.connect(ref_reScale, "out_file", outputnode, "ref")
    workflow.connect(ref_BET, "out_file", outputnode, "ref_brain")
    workflow.connect(ref_BET, "mask_file", outputnode, "ref_mask")

    return workflow
